""" Manager util functions """
import logging

logger = logging.getLogger(__name__)


def callback(result, data_source_name, split_name):
    """Create callback for 'pool.apply_async'"""
    logger.info(f"{data_source_name} has finished created batches for {split_name}!")


def error_callback(exception, an_error_has_occured, data_source_name, split_name):
    """Create error callback for 'pool.apply_async'

    Need to pass in data_source_name rather than rely on data_source_name
    in the outer scope, because otherwise the error message will contain
    the wrong data_source_name (due to stuff happening concurrently!)
    """
    logger.exception(
        f"Exception raised by {data_source_name} whilst creating batches for"
        f" {split_name}\n{exception.__class__.__name__}: {exception}"
    )
    an_error_has_occured.set()



""" Fixtures for tests """
import os
import tempfile
from datetime import datetime, timedelta

import pytest
from nowcasting_datamodel.connection import DatabaseConnection
from nowcasting_datamodel.models import Base_Forecast, Base_PV, GSPYield, Location, LocationSQL

from nowcasting_dataset.time import floor_minutes_dt


@pytest.fixture()
def gsp_yields_and_systems(db_session):
    """Create gsp yields and systems

    gsp systems: One systems
    GSP yields:
        For system 1, gsp yields from 2 hours ago to 8 in the future at 30 minutes intervals
        For system 2: 1 gsp yield at 16.00
    """

    # this pv systems has same coordiantes as the first gsp
    gsp_yield_sqls = []
    locations = []
    for i in range(338):
        location_sql_1: LocationSQL = Location(
            gsp_id=i + 1,
            label=f"GSP_{i+1}",
            installed_capacity_mw=123.0,
        ).to_orm()

        t0_datetime_utc = floor_minutes_dt(datetime.utcnow()) - timedelta(hours=3)

        gsp_yield_sqls = []
        for hour in range(0, 10):
            for minute in range(0, 60, 30):
                datetime_utc = t0_datetime_utc + timedelta(hours=hour - 2, minutes=minute)
                gsp_yield_1 = GSPYield(
                    datetime_utc=datetime_utc,
                    solar_generation_kw=20 + hour + minute,
                ).to_orm()
                gsp_yield_1.location = location_sql_1
                gsp_yield_sqls.append(gsp_yield_1)
                locations.append(location_sql_1)

    # add to database
    db_session.add_all(gsp_yield_sqls + locations)

    db_session.commit()

    return {
        "gsp_yields": gsp_yield_sqls,
        "gs_systems": locations,
    }


@pytest.fixture
def db_connection():
    """Create data connection"""

    with tempfile.NamedTemporaryFile(suffix=".db") as temp:
        url = f"sqlite:///{temp.name}"
        os.environ["DB_URL_PV"] = url
        os.environ["DB_URL"] = url

        connection = DatabaseConnection(url=url, base=Base_PV)
        Base_PV.metadata.create_all(connection.engine)
        Base_Forecast.metadata.create_all(connection.engine)

        yield connection

        Base_PV.metadata.drop_all(connection.engine)
        Base_Forecast.metadata.create_all(connection.engine)


@pytest.fixture(scope="function", autouse=True)
def db_session(db_connection):
    """Creates a new database session for a test."""

    connection = db_connection.engine.connect()
    t = connection.begin()

    with db_connection.get_session() as s:
        s.begin()
        yield s
        s.rollback()

    t.rollback()
    connection.close()



"""Manager live class."""

import logging
import multiprocessing
from datetime import datetime
from functools import partial
from typing import List, Optional

import numpy as np
import pandas as pd

import nowcasting_dataset.utils as nd_utils
from nowcasting_dataset.consts import (
    N_GSPS,
    SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME,
)
from nowcasting_dataset.data_sources.metadata.metadata_model import (
    Metadata,
    SpaceTimeLocation,
    load_from_csv,
)
from nowcasting_dataset.dataset.batch import Batch
from nowcasting_dataset.filesystem import utils as nd_fs_utils
from nowcasting_dataset.manager.base import ManagerBase
from nowcasting_dataset.manager.utils import callback, error_callback

logger = logging.getLogger(__name__)


class ManagerLive(ManagerBase):
    """The Manager initializes and manage a dict of DataSource objects.

    1. Load configuration file
    2. Initialize the data sources
    3. Make Locations
    4. Create batches.
        For GSP there will be 338 examples.
        For a batch size of 32, there will be 11 batches,
        and therefore we upload at the end of making all the batches

    Attrs:
      config: Configuration object.
      data_sources: dict[str, DataSource]
      data_source_which_defines_geospatial_locations: DataSource: The DataSource used to compute the
        geospatial locations of each example.
      save_batches_locally_and_upload: bool: Set to True by `load_yaml_configuration()` if
        `config.process.upload_every_n_batches > 0`.
    """

    def create_files_specifying_spatial_and_temporal_locations_of_each_example(
        self, t0_datetime: datetime, n_gsps: Optional[int] = N_GSPS
    ) -> None:
        """
        Creates CSV files specifying the locations of each example if those files don't exist yet.

        - Creates one file per split, in this location:
        `<output_data.filepath> / 'live' / spatial_and_temporal_locations_of_each_example.csv`
        - Creates the output directory if it does not exist.
        - This works on any compute environment.

        Args:
            t0_datetime: The datetime for the batch
            n_gsps: the number of gsps we want to make, this is defaulted to N_GSPS=338

        Returns: Nothing
        """

        logger.info(
            f"{SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME} does not exist so"
            " will create..."
        )

        split_name = "live"
        datetimes_for_split = [t0_datetime]

        path_for_csv = self.config.output_data.filepath / split_name

        n_batches_requested = int(np.ceil(n_gsps / self.config.process.batch_size))

        logger.debug(
            f"Creating {n_batches_requested:,d} batches x {self.config.process.batch_size:,d}"
            f" examples per batch = {n_batches_requested:,d} examples for {split_name}."
        )

        locations = self.sample_spatial_and_temporal_locations_for_examples(
            t0_datetime=datetimes_for_split[0], n_examples=n_gsps
        )

        metadata = Metadata(
            batch_size=self.config.process.batch_size, space_time_locations=locations
        )
        # output_filename = self._filename_of_locations_csv_file(split_name="live")
        logger.info(f"Making {path_for_csv} if it does not exist.")
        nd_fs_utils.makedirs(path_for_csv, exist_ok=True)
        logger.debug(f"Writing {path_for_csv}")
        metadata.save_to_csv(path_for_csv)

    def sample_spatial_and_temporal_locations_for_examples(
        self, t0_datetime: datetime, n_examples: Optional[int] = None
    ) -> List[SpaceTimeLocation]:
        """
        Computes the geospatial and temporal locations for each training example.

        The first data_source in this DataSourceList defines the geospatial locations of
        each example.

        Args:
            t0_datetimes: All available t0 datetimes.  Can be computed with
                `DataSourceList.get_t0_datetimes_across_all_data_sources()`
            n_examples: The number of examples requested.
            get_all_locations: optional to return all locations for each t0_datetime

        Returns:
            Each row of each the DataFrame specifies the position of each example, using
            columns: 't0_datetime_UTC', 'x_center_OSGB', 'y_center_OSGB'.
        """

        # note that the returned 'shuffled_t0_datetimes'
        # has duplicate datetimes for each location
        locations: List[
            SpaceTimeLocation
        ] = self.data_source_which_defines_geospatial_locations.get_all_locations(
            t0_datetimes_utc=pd.DatetimeIndex([t0_datetime])
        )

        # reduce locations to n_examples
        if n_examples is not None:
            locations = locations[:n_examples]

        # find out the number of examples in the last batch,
        # we maybe need to duplicate the last example into order to get a full batch
        n_examples_last_batch = len(locations) % self.config.process.batch_size
        # Note 0 means the examples fit into the batches

        if n_examples_last_batch != 0:
            # add extra examples to make sure the number of examples fits into batches.
            # There could be a different way to do this,
            # but this keeps it pretty simple for the moment
            extra_examples_needed = self.config.process.batch_size - n_examples_last_batch
            for _ in range(0, extra_examples_needed):
                locations.append(
                    SpaceTimeLocation(
                        t0_datetime_utc=locations[0].t0_datetime_utc,
                        x_center_osgb=locations[0].x_center_osgb,
                        y_center_osgb=locations[0].y_center_osgb,
                        id=locations[0].id,
                    )
                )

        return locations

    def create_batches(self, use_async: Optional[bool] = True) -> None:
        """Create batches (if necessary).

        Make dirs: `<output_data.filepath> / <split_name> / <data_source_name>`.

        Also make `local_temp_path` if necessary.

        """
        logger.debug("Entering Manager.create_batches...")

        # Load locations for each example off disk.
        # locations_for_each_example_of_each_split: dict[split.SplitName, pd.DataFrame] = {}
        # for split_name in splits_which_need_more_batches:
        filename = self._filename_of_locations_csv_file(split_name="live")
        logger.info(f"Loading {filename}.")

        # TODO add pydantic model for this
        metadata = load_from_csv(path=filename, batch_size=self.config.process.batch_size)
        locations_for_each_example = metadata.space_time_locations

        # Fire up a separate process for each DataSource, and pass it a list of batches to
        # create, and whether to utils.upload_and_delete_local_files().
        # TODO: Issue 321: Split this up into separate functions!!!
        n_data_sources = len(self.data_sources)
        nd_utils.set_fsspec_for_multiprocess()

        split_name = "live"
        # TODO Need to think if this is the correct option for live.
        with multiprocessing.Pool(processes=n_data_sources) as pool:
            async_results_from_create_batches = []
            an_error_has_occured = multiprocessing.Event()
            for worker_id, (data_source_name, data_source) in enumerate(self.data_sources.items()):

                # Get indexes of first batch and example. And subset locations_for_split.
                idx_of_first_batch = 0
                locations = locations_for_each_example

                # Get paths.
                dst_path = self.config.output_data.filepath / split_name / data_source_name

                # TODO: Issue 455: Guarantee that local temp path is unique and empty.
                local_temp_path = (
                    self.config.process.local_temp_path
                    / split_name
                    / data_source_name
                    / f"worker_{worker_id}"
                )

                # Make folders.
                nd_fs_utils.makedirs(dst_path, exist_ok=True)
                if dst_path != local_temp_path:
                    nd_fs_utils.makedirs(local_temp_path, exist_ok=True)

                # Key word arguments to be passed into data_source.create_batches():
                kwargs_for_create_batches = dict(
                    spatial_and_temporal_locations_of_each_example=locations,
                    idx_of_first_batch=idx_of_first_batch,
                    batch_size=self.config.process.batch_size,
                    dst_path=dst_path,
                    local_temp_path=local_temp_path,
                    upload_every_n_batches=self.config.process.upload_every_n_batches,
                )

                # Submit data_source.create_batches task to the worker process.
                logger.debug(
                    f"About to submit create_batches task for {data_source_name}, {split_name}"
                )

                if ~use_async:
                    # Sometimes when debuggin it is easy to use non async
                    data_source.create_batches(**kwargs_for_create_batches)
                else:

                    async_result = pool.apply_async(
                        data_source.create_batches,
                        kwds=kwargs_for_create_batches,
                        callback=partial(
                            callback, data_source_name=data_source_name, split_name=split_name
                        ),
                        error_callback=partial(
                            error_callback,
                            data_source_name=data_source_name,
                            split_name=split_name,
                            an_error_has_occured=an_error_has_occured,
                        ),
                    )
                    async_results_from_create_batches.append(async_result)

                    # Wait for all async_results to finish:
                    for async_result in async_results_from_create_batches:
                        async_result.wait()
                        if an_error_has_occured.is_set():
                            # An error has occurred but, at this point in the code,
                            # we don't know which worker process raised the exception.
                            # But, with luck, the worker process
                            # will have logged an informative exception via the
                            # _error_callback func.
                            raise RuntimeError(
                                f"A worker process raised an exception whilst "
                                f"working on {split_name}!"
                            )

            logger.info(f"Finished creating batches for {split_name}!")

    def save_batch(self, batch_idx, path: str):
        """
        Option to save batch to a differnt location

        Args:
            path: the path to save the file to
            batch_idx: the batch index

        """

        # load batch
        batch = Batch.load_netcdf(
            batch_idx=batch_idx,
            data_sources_names=self.data_sources,
            local_netcdf_path=self.config.output_data.filepath / "live",
        )

        # save batch
        batch.save_netcdf(batch_i=batch_idx, path=path)



""" Satellite Data Source """
import itertools
import logging
from dataclasses import InitVar, dataclass
from datetime import datetime, timedelta
from functools import partial
from numbers import Number
from typing import Callable, Iterable, Optional

import dask
import numpy as np
import pandas as pd
import pyproj
import pyresample
import xarray as xr

import nowcasting_dataset.time as nd_time
from nowcasting_dataset.consts import SAT_VARIABLE_NAMES
from nowcasting_dataset.data_sources.data_source import ZarrDataSource
from nowcasting_dataset.data_sources.metadata.metadata_model import SpaceTimeLocation
from nowcasting_dataset.data_sources.satellite.satellite_model import Satellite
from nowcasting_dataset.geospatial import OSGB
from nowcasting_dataset.utils import drop_duplicate_times, drop_non_monotonic_increasing, is_sorted

_LOG = logging.getLogger(__name__)
_LOG_HRV = logging.getLogger(__name__.replace("satellite", "hrvsatellite"))


@dataclass
class SatelliteDataSource(ZarrDataSource):
    """Satellite Data Source."""

    channels: Optional[Iterable[str]] = SAT_VARIABLE_NAMES[1:]
    image_size_pixels_height: InitVar[int] = 128
    image_size_pixels_width: InitVar[int] = 256
    meters_per_pixel: InitVar[int] = 2_000
    logger = _LOG
    time_resolution_minutes: int = 5
    keep_dawn_dusk_hours: int = 0
    is_live: bool = False
    live_delay_minutes: int = 30

    def __post_init__(
        self, image_size_pixels_height: int, image_size_pixels_width: int, meters_per_pixel: int
    ):
        """Post Init"""
        assert len(self.channels) > 0, "channels cannot be empty!"
        assert image_size_pixels_height > 0, "image_size_pixels_height cannot be <= 0!"
        assert image_size_pixels_width > 0, "image_size_pixels_width cannot be <= 0!"
        assert meters_per_pixel > 0, "meters_per_pixel cannot be <= 0!"
        super().__post_init__(image_size_pixels_height, image_size_pixels_width, meters_per_pixel)
        n_channels = len(self.channels)

        if self.is_live:
            # This is to account for the delay in satellite data
            self.total_seq_length = (
                self.history_length
                - int(self.live_delay_minutes / self.time_resolution_minutes)
                + 1
            )

        self._shape_of_example = (
            self.total_seq_length,
            image_size_pixels_height,
            image_size_pixels_width,
            n_channels,
        )

        self._osgb_to_geostationary: Callable = None

    @property
    def sample_period_minutes(self) -> int:
        """Override the default sample minutes"""
        return self.time_resolution_minutes

    def open(self) -> None:
        """
        Open Satellite data

        We don't want to open_sat_data in __init__.
        If we did that, then we couldn't copy SatelliteDataSource
        instances into separate processes.  Instead,
        call open() _after_ creating separate processes.
        """
        self._data = self._open_data()
        if not set(self.channels).issubset(self._data.channels.values):
            raise RuntimeError(
                f"One or more requested channels are not available in {self.zarr_path}!"
                f"  Requested channels={self.channels}."
                f"  Available channels={self._data.channels.values}"
            )
        self._data = self._data.sel(channels=list(self.channels))
        self._load_geostationary_area_definition_and_transform()

        # Check the x and y coords are ascending. If they are not then searchsorted won't work!
        assert is_sorted(self._data.x_geostationary)
        assert is_sorted(self._data.y_geostationary)

    def _load_geostationary_area_definition_and_transform(self) -> None:
        area_definition_yaml = self._data.attrs["area"]
        geostationary_area_definition = pyresample.area_config.load_area_from_string(
            area_definition_yaml
        )
        geostationary_crs = geostationary_area_definition.crs
        self._osgb_to_geostationary = pyproj.Transformer.from_crs(
            crs_from=OSGB, crs_to=geostationary_crs
        ).transform

    def _open_data(self) -> xr.DataArray:
        data_array = open_sat_data(
            zarr_path=self.zarr_path,
            consolidated=self.consolidated,
            logger=self.logger,
            sample_period_minutes=self.sample_period_minutes,
        )
        # If live, then load into memory so the data
        # doesn't disappear while its being used
        if self.is_live:
            data_array = data_array.load()

            # 1 month a year, and 2 days a month the 5 minute satellite data is down,
            # but the 15 minutes data is still there, so lets check if we need to load that

            latest_time = pd.to_datetime(data_array.time[-1].values)
            if latest_time < datetime.utcnow() - timedelta(hours=1):
                self.logger.debug(f"last datesatmp is {latest_time}, which is more than 1 hour ago")
                # create new filename with _15 on the end
                # file name is xxxxxx.zarr or xxxxx.zarr.zip,
                zarr_path_15 = str(self.zarr_path).replace(".zarr", "_15.zarr")

                self.logger.debug(f"Now going to load {zarr_path_15} and resample")
                data_array = open_sat_data(
                    zarr_path=zarr_path_15,
                    consolidated=self.consolidated,
                    logger=self.logger,
                    sample_period_minutes=self.sample_period_minutes,
                )

                data_array = data_array.load()
                data_array = data_array.resample(time="5T").interpolate("linear")

        return data_array

    @staticmethod
    def get_data_model_for_batch():
        """Get the model that is used in the batch"""
        return Satellite

    def _get_time_slice(self, t0_datetime_utc: pd.Timestamp) -> xr.DataArray:
        start_dt = self._get_start_dt(t0_datetime_utc)
        end_dt = self._get_end_dt(t0_datetime_utc)

        # if live data, take 30 ins from the end time,
        # so that we account for delay in data from satellite
        if self.is_live:
            end_dt = end_dt - timedelta(minutes=self.live_delay_minutes)

        # floor to 15 mins
        start_floor = start_dt.floor(f"{self.sample_period_minutes}T")
        end_floor = end_dt.floor(f"{self.sample_period_minutes}T")

        data = self.data.sel(time=slice(start_floor, end_floor))
        assert type(data) == xr.DataArray

        return data

    def get_spatial_region_of_interest(
        self, data_array: xr.DataArray, x_center_osgb: Number, y_center_osgb: Number
    ) -> xr.DataArray:
        """
        Gets the satellite image as a square around the center

        Ignores x and y coordinates as for the original satellite projection each pixel varies in
        both its x and y distance from other pixels. See Issue 401 for more details.

        This results, in 'real' spatial terms, each image covering about 2x as much distance in the
        x direction as in the y direction.

        Args:
            data_array: DataArray to subselect from
            x_center_osgb: Center of the image x coordinate in OSGB coordinates
            y_center_osgb: Center of image y coordinate in OSGB coordinates

        Returns:
            The selected data around the center
        """
        x_center_geostationary, y_center_geostationary = self._osgb_to_geostationary(
            x_center_osgb, y_center_osgb
        )
        # Get the index into x and y nearest to x_center_geostationary and y_center_geostationary:
        x_index_at_center = (
            np.searchsorted(data_array.x_geostationary.values, x_center_geostationary) - 1
        )
        y_index_at_center = (
            np.searchsorted(data_array.y_geostationary.values, y_center_geostationary) - 1
        )
        # Put x_index_at_center and y_index_at_center into a pd.Series so we can operate
        # on them both in a single line of code.
        x_and_y_index_at_center = pd.Series(
            {"x_index_at_center": x_index_at_center, "y_index_at_center": y_index_at_center}
        )
        half_image_size_pixels_height = self._rectangle.size_pixels_height // 2
        half_image_size_pixels_width = self._rectangle.size_pixels_width // 2
        min_x_and_y_index_width = x_and_y_index_at_center - half_image_size_pixels_width
        max_x_and_y_index_width = x_and_y_index_at_center + half_image_size_pixels_width
        min_x_and_y_index_height = x_and_y_index_at_center - half_image_size_pixels_height
        max_x_and_y_index_height = x_and_y_index_at_center + half_image_size_pixels_height

        # Check whether the requested region of interest steps outside of the available data:
        # Need to know how much to pad the outputs, so can do that here
        left_width_padding = max(-min_x_and_y_index_width.x_index_at_center, 0)
        right_width_padding = max(
            max_x_and_y_index_width.x_index_at_center - len(data_array.x_geostationary), 0
        )
        bottom_height_padding = max(-min_x_and_y_index_height.y_index_at_center, 0)
        top_height_padding = max(
            max_x_and_y_index_height.y_index_at_center - len(data_array.y_geostationary), 0
        )

        # Select the geographical region of interest.
        # Note that isel is *exclusive* of the end of the slice.
        # e.g. isel(x=slice(0, 3)) will return the first, second, and third values.
        data_array = data_array.isel(
            x_geostationary=slice(
                max(min_x_and_y_index_width.x_index_at_center, 0),
                min(max_x_and_y_index_width.x_index_at_center, len(data_array.x_geostationary)),
            ),
            y_geostationary=slice(
                max(min_x_and_y_index_height.y_index_at_center, 0),
                min(max_x_and_y_index_height.y_index_at_center, len(data_array.y_geostationary)),
            ),
        )

        # isel is wrong if padding before, so add padding after to be the correct size
        # Get the difference for each direction and use that
        if (
            bottom_height_padding > 0
            or left_width_padding > 0
            or top_height_padding > 0
            or right_width_padding > 0
        ):
            data_array = data_array.pad(
                pad_width={
                    "x_geostationary": (left_width_padding, right_width_padding),
                    "y_geostationary": (bottom_height_padding, top_height_padding),
                },
                mode="constant",
                constant_values=0,
            )

        return data_array

    def get_example(self, location: SpaceTimeLocation) -> xr.Dataset:
        """
        Get Example data

        Args:
            location: A location object of the example which contains
                - a timestamp of the example (t0_datetime_utc),
                - the x center location of the example (x_location_osgb)
                - the y center location of the example(y_location_osgb)

        Returns: Example Data

        """
        t0_datetime_utc = location.t0_datetime_utc
        x_center_osgb = location.x_center_osgb
        y_center_osgb = location.y_center_osgb

        selected_data = self._get_time_slice(t0_datetime_utc)
        selected_data = self.get_spatial_region_of_interest(
            data_array=selected_data,
            x_center_osgb=x_center_osgb,
            y_center_osgb=y_center_osgb,
        )

        selected_data = self._post_process_example(selected_data, t0_datetime_utc)

        if selected_data.shape != self._shape_of_example:
            raise RuntimeError(
                "Example is wrong shape! "
                f"x_center_osgb={x_center_osgb}\n"
                f"y_center_osgb={y_center_osgb}\n"
                f"t0_dt={t0_datetime_utc}\n"
                f"times are {selected_data.time}\n"
                f"expected shape={self._shape_of_example}\n"
                f"actual shape {selected_data.shape}"
                f" {self.forecast_length=}"
                f" {self.history_length=}"
            )

        # Delete the attributes because they fail to save to HDF5.
        # HDF complains with the exception `TypeError: No conversion path for dtype: dtype('<U32')`
        # (If the downstream code wants to use attributes then we can almost certainly find a way
        # to encode the attributes in an HDF5-friendly fashion. The error was
        # See https://docs.h5py.org/en/latest/strings.html#what-about-numpy-s-u-type
        selected_data.attrs = {}

        return selected_data.load().to_dataset(name="data")

    def datetime_index(self, remove_night: bool = True) -> pd.DatetimeIndex:
        """Returns a complete list of all available datetimes

        Args:
            remove_night: If True then remove datetimes at night.
                We're interested in forecasting solar power generation, so we
                don't care about nighttime data :)

                In the UK in summer, the sun rises first in the north east, and
                sets last in the north west [1].  In summer, the north gets more
                hours of sunshine per day.

                In the UK in winter, the sun rises first in the south east, and
                sets last in the south west [2].  In winter, the south gets more
                hours of sunshine per day.

                |                        | Summer | Winter |
                |           ---:         |  :---: |  :---: |
                | Sun rises first in     | N.E.   | S.E.   |
                | Sun sets last in       | N.W.   | S.W.   |
                | Most hours of sunlight | North  | South  |

                Before training, we select timesteps which have at least some
                sunlight.  We do this by computing the clearsky global horizontal
                irradiance (GHI) for the four corners of the satellite imagery,
                and for all the timesteps in the dataset.  We only use timesteps
                where the maximum global horizontal irradiance across all four
                corners is above some threshold.

                The 'clearsky solar irradiance' is the amount of sunlight we'd
                expect on a clear day at a specific time and location. The SI unit
                of irradiance is watt per square meter.  The 'global horizontal
                irradiance' (GHI) is the total sunlight that would hit a
                horizontal surface on the surface of the Earth.  The GHI is the
                sum of the direct irradiance (sunlight which takes a direct path
                from the Sun to the Earth's surface) and the diffuse horizontal
                irradiance (the sunlight scattered from the atmosphere).  For more
                info, see: https://en.wikipedia.org/wiki/Solar_irradiance

        References:
          1. [Video of June 2019](https://www.youtube.com/watch?v=IOp-tj-IJpk)
          2. [Video of Jan 2019](https://www.youtube.com/watch?v=CJ4prUVa2nQ)
        """
        if self._data is None:
            sat_data = self._open_data()
        else:
            sat_data = self._data

        datetime_index = pd.DatetimeIndex(sat_data.time.values)

        if remove_night:
            border_locations = self.geospatial_border()
            datetime_index = nd_time.select_daylight_datetimes(
                datetimes=datetime_index,
                locations=border_locations,
                keep_dawn_dusk_hours=self.keep_dawn_dusk_hours,
            )

        return datetime_index

    def geospatial_border(self) -> list[tuple[Number, Number]]:
        """
        Get 'corner' coordinates for a rectangle within the boundary of the data.

        Returns List of 2-tuples of the x and y coordinates of each corner,
        in OSGB projection.
        """
        GEO_BORDER: int = 64  #: In same geo projection and units as sat_data.
        data = self._open_data()
        return [
            (
                data.x_osgb.isel(x_geostationary=x, y_geostationary=y).values,
                data.y_osgb.isel(x_geostationary=x, y_geostationary=y).values,
            )
            for x, y in itertools.product([GEO_BORDER, -GEO_BORDER], [GEO_BORDER, -GEO_BORDER])
        ]


class HRVSatelliteDataSource(SatelliteDataSource):
    """Satellite Data Source for HRV data."""

    channels: Optional[Iterable[str]] = SAT_VARIABLE_NAMES[:1]
    image_size_pixels: InitVar[int] = 128
    meters_per_pixel: InitVar[int] = 2_000
    logger = _LOG_HRV


def dedupe_time_coords(dataset: xr.Dataset, logger: logging.Logger) -> xr.Dataset:
    """
    Preprocess datasets by de-duplicating the time coordinates.

    Args:
        dataset: xr.Dataset to preprocess
        logger: logger object to write to

    Returns:
        dataset with time coords de-duped.
    """
    # If there are any duplicated init_times then drop the duplicated time:
    data = drop_duplicate_times(data_array=dataset["data"], class_name="Satellite", time_dim="time")

    # If any init_times are not monotonic_increasing then drop the out-of-order init_times:
    data = drop_non_monotonic_increasing(data_array=data, class_name="Satellite", time_dim="time")
    dataset = data.to_dataset(name="data")

    assert pd.DatetimeIndex(data["time"]).is_unique
    assert pd.DatetimeIndex(data["time"]).is_monotonic_increasing

    return dataset


def open_sat_data(
    zarr_path: str, consolidated: bool, logger: logging.Logger, sample_period_minutes: int = 15
) -> xr.DataArray:
    """Lazily opens the Zarr store.

    Adds 1 minute to the 'time' coordinates, so the timestamps
    are at 00, 05, ..., 55 past the hour.

    Args:
      zarr_path: Cloud URL or local path pattern.  If GCP URL, must start with 'gs://'
      consolidated: Whether or not the Zarr metadata is consolidated.
      logger: logger object to write to
      sample_period_minutes: The sample period minutes that the data should be reduced to.
    """
    logger.debug("Opening satellite data: %s", zarr_path)

    # If we are opening multiple Zarr stores (i.e. one for each month of the year) we load them
    # together and create a single dataset from them.  open_mfdataset also works if zarr_path
    # points to a specific zarr directory (with no wildcards).

    # Silence the warning about large chunks.
    # Alternatively, we could set this to True, but that slows down loading a Satellite batch
    # from 8 seconds to 50 seconds!
    dask.config.set(**{"array.slicing.split_large_chunks": False})

    # add logger to preprocess function
    p_dedupe_time_coords = partial(dedupe_time_coords, logger=logger)
    if str(zarr_path).split(".")[-1] == "zip":
        if "s3://" or "gs://" in str(zarr_path):
            data_path = "zip::" + str(zarr_path)
        else:
            data_path = "zip://" + str(zarr_path)
        dataset = xr.open_dataset(data_path, engine="zarr")

    else:
        # Open datasets.
        dataset = xr.open_mfdataset(
            zarr_path,
            chunks="auto",  # See issue #456 for why we use "auto".
            mode="r",
            engine="zarr",
            concat_dim="time",
            preprocess=p_dedupe_time_coords,
            consolidated=consolidated,
            combine="nested",
        )

    # Rename
    # These renamings will no longer be necessary when the Zarr uses the 'correct' names,
    # see https://github.com/openclimatefix/Satip/issues/66
    if "x" in dataset:
        dataset = dataset.rename({"x": "x_geostationary", "y": "y_geostationary"})
    if "variable" in dataset:
        dataset = dataset.rename({"variable": "channels"})
    elif "channels" not in dataset:
        # This is HRV version 3, which doesn't have a channels dim.  So add one.
        dataset = dataset.expand_dims(dim={"channels": ["HRV"]}, axis=-1)

    data_array = dataset["data"]
    if "stacked_eumetsat_data" == data_array.name:
        data_array.name = "data"
    del dataset

    # Flip coordinates to top-left first
    data_array = data_array.reindex(x_geostationary=data_array.x_geostationary[::-1])

    # reindex satellite to 15 mins data
    datetime_index = pd.DatetimeIndex(data_array["time"])
    if sample_period_minutes != 5:
        time_mask = datetime_index.minute % sample_period_minutes == 0
        data_array = data_array.sel(time=time_mask)

    # Sanity check!
    assert datetime_index.is_unique
    assert datetime_index.is_monotonic_increasing

    return data_array



# noqa: D100
import pandas as pd

from nowcasting_dataset import utils


def test_is_monotically_increasing():  # noqa: D103
    assert utils.is_monotonically_increasing([1, 2, 3, 4])
    assert not utils.is_monotonically_increasing([1, 2, 3, 3])
    assert not utils.is_monotonically_increasing([1, 2, 3, 0])

    index = pd.date_range("2010-01-01", freq="H", periods=4)
    assert utils.is_monotonically_increasing(index)
    assert not utils.is_monotonically_increasing(index[::-1])


def test_drop_monotonically_increasing_nothing():
    """Test to check nothing is dropped when times are correctly in order"""
    times = ["2022-01-01", "2022-01-02", "2022-01-03"]
    data_xr = pd.Series(index=pd.DatetimeIndex(times), data=[1, 2, 3]).to_xarray()

    t = utils.drop_non_monotonic_increasing(
        data_array=data_xr, class_name="test_class", time_dim="index"
    )
    assert (t == data_xr).all()


def test_drop_monotonically_increasing():
    """Test to correct value is droped when times are not in order"""
    times = ["2022-01-02", "2022-01-01", "2022-01-03"]
    data_xr = pd.Series(index=pd.DatetimeIndex(times), data=[1, 2, 3]).to_xarray()

    t = utils.drop_non_monotonic_increasing(
        data_array=data_xr, class_name="test_class", time_dim="index"
    )
    assert (t.index == pd.DatetimeIndex(["2022-01-01", "2022-01-03"])).all()


def test_drop_duplicate_nothing():
    """Test to check nothing is dropped when times are not duplicated"""
    times = ["2022-01-01", "2022-01-02", "2022-01-03"]
    data_xr = pd.Series(index=pd.DatetimeIndex(times), data=[1, 2, 3]).to_xarray()

    t = utils.drop_duplicate_times(data_array=data_xr, class_name="test_class", time_dim="index")
    assert (t == data_xr).all()


def test_drop_duplicate():
    """Test to correct value is droped when times are duplicated"""
    times = ["2022-01-01", "2022-01-01", "2022-01-03"]
    data_xr = pd.Series(index=pd.DatetimeIndex(times), data=[1, 2, 3]).to_xarray()

    t = utils.drop_duplicate_times(data_array=data_xr, class_name="test_class", time_dim="index")
    assert (t.index == pd.DatetimeIndex(["2022-01-01", "2022-01-03"])).all()


def test_get_netcdf_filename():  # noqa: D103
    assert utils.get_netcdf_filename(10) == "000010.nc"


def test_remove_regex_pattern_from_keys():  # noqa: D103
    d = {
        "satellite_zarr_path": "/a/b/c/foo.zarr",
        "bar": "baz",
        "satellite_channels": ["HRV"],
        "n_satellite_per_batch": 4,
    }
    correct = {
        "zarr_path": "/a/b/c/foo.zarr",
        "bar": "baz",
        "channels": ["HRV"],
        "n_satellite_per_batch": 4,
    }
    new_dict = utils.remove_regex_pattern_from_keys(d, pattern_to_remove=r"^satellite_")
    assert new_dict == correct



""" Model for output of Optical Flow data """
from __future__ import annotations

from nowcasting_dataset.data_sources.datasource_output import DataSourceOutput


class OpticalFlow(DataSourceOutput):
    """Class to store optical flow data as a xr.Dataset with some validation"""

    __slots__ = ()
    _expected_dimensions = ("time", "y_geostationary", "x_geostationary", "channels")
    _expected_data_vars = ("data",)

    @classmethod
    def model_validation(cls, v):
        """Check that all values are not NaN, Infinite, or -1."""
        v.check_nan_and_inf(data=v.data)
        # previously nans were filled with -1s, so lets make sure there are none
        v.check_dataset_not_equal(data=v.data, value=-1)
        v.check_data_var_dim(v.x_geostationary, ("example", "x_geostationary_index"))
        v.check_data_var_dim(v.y_geostationary, ("example", "y_geostationary_index"))

        return v



""" PV Data Source """

import datetime
import functools
import logging
from dataclasses import dataclass
from numbers import Number
from pathlib import Path
from typing import List, Optional, Tuple, Union

import fsspec
import numpy as np
import pandas as pd
import xarray as xr

import nowcasting_dataset.filesystem.utils as nd_fs_utils
from nowcasting_dataset import geospatial
from nowcasting_dataset.config.model import PVFiles
from nowcasting_dataset.consts import DEFAULT_N_PV_SYSTEMS_PER_EXAMPLE
from nowcasting_dataset.data_sources.data_source import ImageDataSource
from nowcasting_dataset.data_sources.metadata.metadata_model import SpaceTimeLocation
from nowcasting_dataset.data_sources.pv.live import (
    get_metadata_from_database,
    get_pv_power_from_database,
)
from nowcasting_dataset.data_sources.pv.pv_model import PV
from nowcasting_dataset.data_sources.pv.utils import encode_label
from nowcasting_dataset.square import get_bounding_box_mask, get_closest_coordinate_order

logger = logging.getLogger(__name__)


@dataclass
class PVDataSource(ImageDataSource):
    """PV Data Source.

    This inherits from ImageDataSource so PVDataSource can select a geospatial region of interest
    defined by image_size_pixels and meters_per_pixel.
    """

    files_groups: List[Union[PVFiles, dict]]
    # TODO: Issue #425: Use config to set start_dt and end_dt.
    start_datetime: Optional[datetime.datetime] = None
    end_datetime: Optional[datetime.datetime] = None
    random_pv_system_for_given_location: Optional[bool] = True
    #: Each example will always have this many PV systems.
    #: If less than this number exist in the data then pad with NaNs.
    n_pv_systems_per_example: int = DEFAULT_N_PV_SYSTEMS_PER_EXAMPLE
    load_azimuth_and_elevation: bool = False
    load_from_gcs: bool = True  # option to load data from gcs, or local file
    get_center: bool = True
    is_live: bool = False
    live_interpolate_minutes: int = 30
    live_load_extra_minutes: int = 60

    def __post_init__(
        self, image_size_pixels_height: int, image_size_pixels_width: int, meters_per_pixel: int
    ):
        """Post Init"""

        if type(self.files_groups[0]) == dict:
            self.files_groups = [PVFiles(**files) for files in self.files_groups]

        super().__post_init__(image_size_pixels_height, image_size_pixels_width, meters_per_pixel)

        self.rng = np.random.default_rng()
        self.load()

    def check_input_paths_exist(self) -> None:
        """Check input paths exist.

        If using live data, dont check Files, as we are loading data from the database
        If not, raise a FileNotFoundError.
        """
        if self.is_live:
            return None

        for pv_files in self.files_groups:
            for filename in [pv_files.pv_filename, pv_files.pv_metadata_filename]:
                nd_fs_utils.check_path_exists(filename)

    def load(self):
        """
        Load metadata and pv power
        """
        self._load_metadata()
        self._load_pv_power()

        self.pv_metadata, self.pv_power = align_pv_system_ids(self.pv_metadata, self.pv_power)

    @staticmethod
    def get_data_model_for_batch():
        """Get the model that is used in the batch"""
        return PV

    def _load_metadata(self):

        logger.debug(f"Loading PV metadata from {self.files_groups}")

        # collect all metadata together
        if not self.is_live:
            pv_metadata = []
            for pv_files in self.files_groups:
                metadata_filename = pv_files.pv_metadata_filename

                # read metadata file
                metadata = pd.read_csv(metadata_filename, index_col="system_id")

                # encode index, to make sure the indexes are unique
                metadata.index = encode_label(indexes=metadata.index, label=pv_files.label)

                # filter for zero capacity
                if pv_files.label == "pvoutput":
                    metadata = metadata[metadata["system_size_watts"] > 0]
                elif pv_files.label == "passiv":
                    metadata = metadata[metadata["kwp"] > 0]

                pv_metadata.append(metadata)
            pv_metadata = pd.concat(pv_metadata)

        else:
            pv_metadata = get_metadata_from_database()

            logger.debug(f"Found {len(pv_metadata)} pv systems from database")

        # drop any systems with no lon or lat
        pv_metadata.dropna(subset=["longitude", "latitude"], how="any", inplace=True)

        logger.debug(f"Found {len(pv_metadata)} pv systems with locations")

        pv_metadata["location_x"], pv_metadata["location_y"] = geospatial.lat_lon_to_osgb(
            pv_metadata["latitude"], pv_metadata["longitude"]
        )

        # Remove PV systems outside the geospatial boundary of the
        # satellite data:
        GEO_BOUNDARY_OSGB = {
            "WEST": -238_000,
            "EAST": 856_000,
            "NORTH": 1_222_000,
            "SOUTH": -184_000,
        }

        self.pv_metadata = pv_metadata[
            (pv_metadata.location_x >= GEO_BOUNDARY_OSGB["WEST"])
            & (pv_metadata.location_x <= GEO_BOUNDARY_OSGB["EAST"])
            & (pv_metadata.location_y <= GEO_BOUNDARY_OSGB["NORTH"])
            & (pv_metadata.location_y >= GEO_BOUNDARY_OSGB["SOUTH"])
        ]

        logger.debug(f"Found {len(pv_metadata)} pv systems")

    def _load_pv_power(self):

        logger.debug(f"Loading PV Power data from {self.files_groups}")

        if not self.is_live:
            # collect all PV power timeseries together
            pv_power_all = []
            for pv_files in self.files_groups:
                filename = pv_files.pv_filename

                # get pv power data
                pv_power = load_solar_pv_data(
                    filename, start_dt=self.start_datetime, end_dt=self.end_datetime
                )

                # encode index, to make sure the columns are unique
                new_columns = encode_label(indexes=pv_power.columns, label=pv_files.label)
                pv_power.columns = new_columns

                pv_power_all.append(pv_power)

            pv_power = pd.concat(pv_power_all, axis="columns")

        else:
            pv_power = get_pv_power_from_database(
                history_duration=self.history_duration,
                interpolate_minutes=self.live_interpolate_minutes,
                load_extra_minutes=self.live_load_extra_minutes,
            )
            logger.debug(f"Found {len(pv_power)} pv power datetimes from database ")
            logger.debug(f"Found {len(pv_power.columns)} pv power pv system ids from database")

        assert not pv_power.columns.duplicated().any()

        # A bit of hand-crafted cleaning
        bad_pvputput_indexes = [30248]
        bad_pvputput_indexes = encode_label(bad_pvputput_indexes, label="pvoutput")
        for bad_index in bad_pvputput_indexes:
            if bad_index in pv_power.columns:
                pv_power[bad_index]["2018-10-29":"2019-01-03"] = np.NaN

        # Drop columns and rows with all NaNs.
        pv_power.dropna(axis="columns", how="all", inplace=True)
        pv_power.dropna(axis="index", how="all", inplace=True)

        logger.debug(f"Found {len(pv_power)} pv power datetimes from database ")
        logger.debug(f"Found {len(pv_power.columns)} pv power pv system ids from database")

        # drop systems with over night power
        # TODO: Fix bug!
        # drop_pv_systems_which_produce_overnight expects normalised data,
        # but it doesn't receive normalised data. See Issue #664.
        pv_power = drop_pv_systems_which_produce_overnight(pv_power)

        logger.debug(
            f"Found {len(pv_power)} pv power datetimes " f"after filtering for over night power  "
        )
        logger.debug(
            f"Found {len(pv_power.columns)} pv power pv system ids "
            f"after filtering for over night power"
        )

        # Resample to 5-minutely and interpolate up to 15 minutes ahead.
        # TODO: Issue #301: Give users the option to NOT resample (because Perceiver IO
        # doesn't need all the data to be perfectly aligned).
        if len(pv_power) > 0:
            pv_power = pv_power.resample("5T").interpolate(method="time", limit=3)
            pv_power.dropna(axis="index", how="all", inplace=True)
        # self.pv_power = dd.from_pandas(pv_power, npartitions=3)
        print("pv_power = {:,.1f} MB".format(pv_power.values.nbytes / 1e6))
        self.pv_power = pv_power

        logger.debug(f"Found {len(self.pv_power)} pv power datetimes")
        logger.debug(f"Found {len(self.pv_power.columns)} pv power pv system ids")

        # get the max generation / capacity for each system
        if not self.is_live:
            self.pv_capacity = pv_power.max()
        else:
            self.pv_capacity = self.pv_metadata["installed_capacity_kw"]

    def _get_time_slice(self, t0_datetime_utc: pd.Timestamp) -> tuple[pd.DataFrame]:
        # TODO: Cache this?
        start_dt = self._get_start_dt(t0_datetime_utc)
        end_dt = self._get_end_dt(t0_datetime_utc)

        selected_pv_power = self.pv_power.loc[start_dt:end_dt].dropna(axis="columns", how="any")
        selected_pv_capacity = self.pv_capacity[selected_pv_power.columns]

        pv_power_zero_or_above_flag = selected_pv_power.ge(0).all()

        if pv_power_zero_or_above_flag.sum() != len(selected_pv_power.columns):
            n = len(selected_pv_power.columns) - pv_power_zero_or_above_flag.sum()
            logger.debug(f"Will be removing {n} pv systems as they have negative values")

        selected_pv_power = selected_pv_power.loc[:, pv_power_zero_or_above_flag]
        selected_pv_capacity = selected_pv_capacity.loc[pv_power_zero_or_above_flag]

        # is live, and no values, add a row of nans
        if (len(selected_pv_power) == 0) and self.is_live:
            selected_pv_power = selected_pv_power.append([pd.DataFrame(index=[t0_datetime_utc])])
            selected_pv_capacity = selected_pv_capacity.append([pd.Series(index=[t0_datetime_utc])])

        return selected_pv_power, selected_pv_capacity

    def _get_central_pv_system_id(
        self,
        x_center_osgb: Number,
        y_center_osgb: Number,
        pv_system_ids_with_data_for_timeslice: pd.Index,
    ) -> int:
        # If x_center_osgb and y_center_osgb have been chosen
        # by PVDataSource.pick_locations_for_batch() then we just have
        # to find the pv_system_ids at that exact location.  This is
        # super-fast (a few hundred microseconds).  We use np.isclose
        # instead of the equality operator because floats.
        pv_system_ids_at_location = self.pv_metadata.index[
            np.isclose(self.pv_metadata.location_x, x_center_osgb)
            & np.isclose(self.pv_metadata.location_y, y_center_osgb)
        ]
        if len(pv_system_ids_at_location) == 0:
            # TODO: Implement finding PV systems closest to x_center_osgb,
            # y_center_osgb.  This will probably be quite slow, so always
            # try finding an exact match first (which is super-fast).
            raise NotImplementedError(
                "Not yet implemented the ability to find PV systems *nearest*"
                " (but not at the identical location to) x_center_osgb and"
                " y_center_osgb."
            )

        pv_system_ids = pv_system_ids_with_data_for_timeslice.intersection(
            pv_system_ids_at_location
        )
        if len(pv_system_ids) == 0:
            logger.debug(pv_system_ids_at_location)
            logger.debug(pv_system_ids_with_data_for_timeslice)
            assert len(pv_system_ids) > 0

        # Select just one PV system (the locations in PVOutput.org are quite
        # approximate, so it's quite common to have multiple PV systems
        # at the same nominal lat lon).
        if self.random_pv_system_for_given_location:
            pv_system_id = self.rng.choice(pv_system_ids)
        else:
            pv_system_id = pv_system_ids[0]

        return pv_system_id

    def _get_all_pv_system_ids_in_roi(
        self,
        x_center_osgb: Number,
        y_center_osgb: Number,
        pv_system_ids_with_data_for_timeslice: pd.Index,
    ) -> pd.Index:
        """
        Find the PV system IDs. for all the PV systems within the geospatial

        This is for all the PV systems within the geospatial
        region of interest, defined by self.square.
        """
        logger.debug(f"Getting PV example data for {x_center_osgb} and {y_center_osgb}")

        bounding_box = self._rectangle.bounding_box_centered_on(
            x_center_osgb=x_center_osgb, y_center_osgb=y_center_osgb
        )
        x = self.pv_metadata.location_x
        y = self.pv_metadata.location_y

        # make mask of pv system ids
        mask = get_bounding_box_mask(bounding_box, x, y)
        pv_system_ids = self.pv_metadata.index[mask]
        x = self.pv_metadata.location_x[mask]
        y = self.pv_metadata.location_y[mask]

        # order the pv systems
        mask_order = get_closest_coordinate_order(
            x_center=x_center_osgb, y_center=y_center_osgb, x=x, y=y
        )
        mask_order = mask_order.sort_values()
        pv_system_ids = mask_order.index

        pv_system_ids = pv_system_ids_with_data_for_timeslice.intersection(pv_system_ids)

        # there may not be any pv systems in a GSP region
        # assert len(pv_system_ids) > 0

        return pv_system_ids

    def get_example(self, location: SpaceTimeLocation) -> xr.Dataset:
        """
        Get Example data for PV data

        Args:
            location: A location object of the example which contains
                - a timestamp of the example (t0_datetime_utc),
                - the x center location of the example (x_location_osgb)
                - the y center location of the example(y_location_osgb)

        Returns: Example data

        """
        logger.debug("Getting PV example data")

        t0_datetime_utc = location.t0_datetime_utc
        x_center_osgb = location.x_center_osgb
        y_center_osgb = location.y_center_osgb

        selected_pv_power, selected_pv_capacity = self._get_time_slice(t0_datetime_utc)
        all_pv_system_ids = self._get_all_pv_system_ids_in_roi(
            x_center_osgb, y_center_osgb, selected_pv_power.columns
        )
        if self.get_center:
            central_pv_system_id = self._get_central_pv_system_id(
                x_center_osgb, y_center_osgb, selected_pv_power.columns
            )

            # By convention, the 'target' PV system ID (the one in the center
            # of the image) must be in the first position of the returned arrays.
            all_pv_system_ids = all_pv_system_ids.drop(central_pv_system_id)
            all_pv_system_ids = all_pv_system_ids.insert(loc=0, item=central_pv_system_id)

        all_pv_system_ids = all_pv_system_ids[: self.n_pv_systems_per_example]

        selected_pv_power = selected_pv_power[all_pv_system_ids]
        selected_pv_capacity = selected_pv_capacity[all_pv_system_ids]

        # this provides an index of what pv systesm are in the examples
        pv_system_row_number = np.flatnonzero(self.pv_metadata.index.isin(all_pv_system_ids))
        pv_system_x_coords = self.pv_metadata.location_x[all_pv_system_ids]
        pv_system_y_coords = self.pv_metadata.location_y[all_pv_system_ids]

        # Save data into the PV object...

        # convert to data array
        da = xr.DataArray(
            data=selected_pv_power.values,
            dims=["time", "id"],
            coords=dict(
                id=all_pv_system_ids.values.astype(int),
                time=selected_pv_power.index.values,
            ),
        )

        capacity = xr.DataArray(
            data=selected_pv_capacity.values,
            dims=["id"],
            coords=dict(
                id=all_pv_system_ids.values.astype(int),
            ),
        )

        # convert to dataset
        pv = da.to_dataset(name="power_mw")
        pv["capacity_mwp"] = capacity

        # add pv x coords
        x_coords = xr.DataArray(
            data=pv_system_x_coords.values,
            dims=["id"],
        )

        y_coords = xr.DataArray(
            data=pv_system_y_coords.values,
            dims=["id"],
        )
        pv_system_row_number = xr.DataArray(
            data=pv_system_row_number,
            dims=["id"],
        )
        pv["x_osgb"] = x_coords.astype("float32")
        pv["y_osgb"] = y_coords.astype("float32")
        pv["pv_system_row_number"] = pv_system_row_number

        # pad out so that there are always n_pv_systems_per_example, pad with zeros
        pad_n = self.n_pv_systems_per_example - len(pv.id)
        pv = pv.pad(id=(0, pad_n), power_mw=((0, 0), (0, pad_n)), constant_values=0)

        # format id
        pv.__setitem__("id", pv.id.astype("int32"))

        return pv

    def get_locations(self, t0_datetimes_utc: pd.DatetimeIndex) -> List[SpaceTimeLocation]:
        """Find a valid geographical location for each t0_datetime.

        Returns:  x_locations, y_locations. Each has one entry per t0_datetime.
            Locations are in OSGB coordinates.
        """
        # Set this up as a separate function, so we can cache the result!
        @functools.cache  # functools.cache requires Python >= 3.9
        def _get_pv_system_ids(t0_datetime: pd.Timestamp) -> pd.Int64Index:
            start_dt = self._get_start_dt(t0_datetime)
            end_dt = self._get_end_dt(t0_datetime)
            available_pv_data = self.pv_power.loc[start_dt:end_dt]
            columns_mask = available_pv_data.notna().all().values
            pv_system_ids = available_pv_data.columns[columns_mask]
            assert len(pv_system_ids) > 0
            return pv_system_ids

        # Pick a random PV system for each t0_datetime, and then grab
        # their geographical location.
        locations = []
        for t0_datetime in t0_datetimes_utc:
            pv_system_ids = _get_pv_system_ids(t0_datetime)
            pv_system_id = self.rng.choice(pv_system_ids)

            # Get metadata for PV system
            metadata_for_pv_system = self.pv_metadata.loc[pv_system_id]

            locations.append(
                SpaceTimeLocation(
                    t0_datetime_utc=t0_datetime,
                    x_center_osgb=metadata_for_pv_system.location_x,
                    y_center_osgb=metadata_for_pv_system.location_y,
                    id=pv_system_id,
                    id_type="pv_system",
                )
            )

        # sort locations by id
        locations = sorted(locations, key=lambda d: d.id)

        return locations

    def datetime_index(self) -> pd.DatetimeIndex:
        """Returns a complete list of all available datetimes."""
        return self.pv_power.index


def load_solar_pv_data(
    filename: Union[str, Path],
    start_dt: Optional[datetime.datetime] = None,
    end_dt: Optional[datetime.datetime] = None,
) -> pd.DataFrame:
    """
    Load solar pv data from any compute environment.

    Args:
        filename: filename of file to be loaded
        start_dt: the start datetime, which to trim the data to
        end_dt: the end datetime, which to trim the data to

    Returns: Solar PV data
    """
    logger.debug(f"Loading Solar PV Data from {filename} from {start_dt} to {end_dt}.")

    with fsspec.open(filename, mode="rb") as file:
        pv_power = xr.open_dataset(file, engine="h5netcdf")
        pv_power = pv_power.sel(datetime=slice(start_dt, end_dt))
        pv_power_df = pv_power.to_dataframe()

    # Process the data a little
    pv_power_df = pv_power_df.dropna(axis="columns", how="all")
    pv_power_df = pv_power_df.clip(lower=0, upper=5e7)
    # Convert the pv_system_id column names from strings to ints:
    pv_power_df.columns = [np.int32(col) for col in pv_power_df.columns]

    if "passiv" not in filename:
        pv_power_df = pv_power_df.tz_localize("Europe/London").tz_convert("UTC").tz_convert(None)

    logger.debug("Loading Solar PV Data: done")

    return pv_power_df


def align_pv_system_ids(
    pv_metadata: pd.DataFrame, pv_power: pd.DataFrame
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Only pick PV systems for which we have metadata."""
    pv_system_ids = pv_metadata.index.intersection(pv_power.columns)
    pv_system_ids = np.sort(pv_system_ids)
    pv_power = pv_power[pv_system_ids]
    pv_metadata = pv_metadata.loc[pv_system_ids]
    return pv_metadata, pv_power


def drop_pv_systems_which_produce_overnight(pv_power: pd.DataFrame) -> pd.DataFrame:
    """Drop systems which produce power over night.

    Args:
        pv_power: Normalised to [0, 1].
    """

    if len(pv_power) == 0:
        return pv_power

    # TODO: Of these bad systems, 24647, 42656, 42807, 43081, 51247, 59919
    # might have some salvagable data?
    NIGHT_YIELD_THRESHOLD = 0.4
    night_hours = [22, 23, 0, 1, 2]
    pv_data_at_night = pv_power.loc[pv_power.index.hour.isin(night_hours)]
    pv_above_threshold_at_night = (pv_data_at_night > NIGHT_YIELD_THRESHOLD).any()
    bad_systems = pv_power.columns[pv_above_threshold_at_night]
    logger.debug(f"{len(bad_systems)} bad PV systems found and removed!")
    return pv_power.drop(columns=bad_systems)



""" Function to get data from live database """
import logging
import os
from datetime import datetime, timedelta, timezone
from typing import List, Optional

import pandas as pd
from nowcasting_datamodel.connection import DatabaseConnection
from nowcasting_datamodel.models.base import Base_PV
from nowcasting_datamodel.models.pv import (
    PVSystem,
    PVSystemSQL,
    PVYield,
    PVYieldSQL,
    pv_output,
    solar_sheffield_passiv,
)
from nowcasting_datamodel.read.read_pv import get_pv_systems, get_pv_yield

from nowcasting_dataset.data_sources.pv.utils import encode_label

logger = logging.getLogger(__name__)


def get_metadata_from_database() -> pd.DataFrame:
    """
    Get metadata from database

    Returns: pandas data frame with the following columns
        - latitude
        - longitude
        - kwp
        The index is the pv_system_id
    """

    # make database connection
    url = os.getenv("DB_URL_PV")
    db_connection = DatabaseConnection(url=url, base=Base_PV)

    pv_system_all_df = []
    for provider in [pv_output, solar_sheffield_passiv]:

        with db_connection.get_session() as session:
            # read pv systems
            pv_systems: List[PVSystemSQL] = get_pv_systems(session=session, provider=provider)

            # format locations
            pv_systems_df = pd.DataFrame(
                [(PVSystem.from_orm(pv_system)).__dict__ for pv_system in pv_systems]
            )

        if len(pv_systems_df) == 0:
            return pd.DataFrame(
                columns=["pv_system_id", "latitude", "longitude", "installed_capacity_kw"]
            )

        if provider == pv_output:
            label = "pvoutput"
        else:
            label = "passiv"
        pv_systems_df.index = encode_label(pv_systems_df["pv_system_id"], label=label)
        pv_systems_df = pv_systems_df[["latitude", "longitude", "installed_capacity_kw"]]

        pv_system_all_df.append(pv_systems_df)

    pv_system_all_df = pd.concat(pv_system_all_df)

    return pv_system_all_df


def get_pv_power_from_database(
    history_duration: timedelta,
    interpolate_minutes: int,
    load_extra_minutes: int,
    load_extra_minutes_and_keep: Optional[int] = 30,
) -> pd.DataFrame:
    """
    Get pv power from database

    Args:
        history_duration: a timedelta of how many minutes to load in the past
        interpolate_minutes: how many minutes we should interpolate the data froward for
        load_extra_minutes: the extra minutes we should load, in order to load more data.
            This is because some data from a site lags significantly behind 'now'.
            These extra minutes are not kept but used to interpolate results.
        load_extra_minutes_and_keep: extra minutes to load, but also keep this data.

    Returns:pandas data frame with the following columns pv systems indexes
    The index is the datetime

    """

    logger.info("Loading PV data from database")
    logger.debug(f"{history_duration=} {interpolate_minutes=} {load_extra_minutes=}")

    extra_duration = timedelta(minutes=load_extra_minutes)
    now = pd.to_datetime(datetime.now(tz=timezone.utc)).ceil("5T")
    start_utc = now - history_duration - timedelta(minutes=load_extra_minutes_and_keep)
    start_utc_extra = start_utc - extra_duration

    # create empty dataframe with 5 mins periods
    empty_df = pd.DataFrame(index=pd.date_range(start=start_utc_extra, end=now, freq="5T"))

    # make database connection
    url = os.getenv("DB_URL_PV")
    db_connection = DatabaseConnection(url=url, base=Base_PV)

    with db_connection.get_session() as session:
        pv_yields: List[PVYieldSQL] = get_pv_yield(
            session=session, start_utc=start_utc_extra, correct_data=True
        )

        logger.debug(f"Found {len(pv_yields)} PV yields from the database")

        pv_yields_df = pd.DataFrame(
            [(PVYield.from_orm(pv_yield)).__dict__ for pv_yield in pv_yields]
        )

    if len(pv_yields_df) == 0:
        logger.warning("Found no pv yields, this might cause an error")
    else:
        logger.debug(f"Found {len(pv_yields_df)} pv yields")

    if len(pv_yields_df) == 0:
        return pd.DataFrame(columns=["pv_system_id"])

        # get the system id from 'pv_system_id=xxxx provider=.....'
    pv_yields_df["pv_system_id"] = (
        pv_yields_df["pv_system"].astype(str).str.split(" ").str[0].str.split("=").str[-1]
    )

    pv_yields_df["provider"] = (
        pv_yields_df["pv_system"]
        .astype(str)
        .str.split(" ")
        .str[1]
        .str.split("=")
        .str[1]
        .str.replace("'", "")
    )

    # encode pv system id
    for provider in pv_output, solar_sheffield_passiv:
        idx = pv_yields_df["provider"] == provider
        if provider == pv_output:
            label = "pvoutput"
        else:
            label = "passiv"
        pv_yields_df.loc[idx, "pv_system_id"] = encode_label(
            pv_yields_df.loc[idx, "pv_system_id"], label=label
        )

    # pivot on
    pv_yields_df = pv_yields_df[["datetime_utc", "pv_system_id", "solar_generation_kw"]]
    pv_yields_df.drop_duplicates(
        ["datetime_utc", "pv_system_id", "solar_generation_kw"], keep="last", inplace=True
    )
    pv_yields_df = pv_yields_df.pivot(
        index="datetime_utc", columns="pv_system_id", values="solar_generation_kw"
    )

    # we are going interpolate using 'quadratic' method and we need at least 3 data points,
    # therefore we drop system wiht less than 3 nans
    pv_yields_df = pv_yields_df.loc[:, pv_yields_df.notnull().sum() >= 3]

    # interpolate in between, maximum 'live_interpolate_minutes' mins
    # note data is in 5 minutes chunks
    pv_yields_df = empty_df.join(pv_yields_df)
    limit = int(interpolate_minutes / 5)
    if limit > 0:
        try:
            pv_yields_df.interpolate(limit=limit, inplace=True, method="quadratic")
        except Exception as e:
            logger.exception(e)
            logger.debug(pv_yields_df)
            raise Exception(f"Could not do interpolate with limit {limit}")

    # filter out the extra minutes loaded
    logger.debug(f"{len(pv_yields_df)} of datetimes before filter on {start_utc}")
    pv_yields_df = pv_yields_df[pv_yields_df.index >= start_utc]
    logger.debug(f"{len(pv_yields_df)} of datetimes after filter on {start_utc}")

    return pv_yields_df



""" Function to get data from live database """
import logging
import os
from datetime import datetime, timedelta, timezone
from typing import List

import pandas as pd
from nowcasting_datamodel.connection import DatabaseConnection
from nowcasting_datamodel.models.base import Base_Forecast
from nowcasting_datamodel.models.gsp import GSPYield, GSPYieldSQL, Location
from nowcasting_datamodel.read.read_gsp import get_gsp_yield

from nowcasting_dataset.consts import N_GSPS

logger = logging.getLogger(__name__)


def get_gsp_power_from_database(
    history_duration: timedelta, interpolate_minutes: int, load_extra_minutes: int
) -> (pd.DataFrame, pd.DataFrame):
    """
    Get gsp power from database

    Args:
        history_duration: a timedelta of how many minutes to load in the past
        interpolate_minutes: how many minutes we should interpolate the data froward for
        load_extra_minutes: the extra minutes we should load, in order to load more data.
            This is because some data from a site lags significantly behind 'now'

    Returns:pandas data frame with the following columns pv systems indexes
    The index is the datetime

    """

    logger.info("Loading GSP data from database")
    logger.debug(f"{history_duration=}")

    extra_duration = timedelta(minutes=load_extra_minutes)
    now = pd.to_datetime(datetime.now(tz=timezone.utc)).floor("30T")
    start_utc = now - history_duration
    start_utc_extra = start_utc - extra_duration

    # create empty dataframe with 30 mins periods
    empty_df = pd.DataFrame(
        index=pd.date_range(start=start_utc_extra, end=now, freq="30T", tz=timezone.utc)
    )

    # make database connection
    url = os.getenv("DB_URL")
    db_connection = DatabaseConnection(url=url, base=Base_Forecast)

    with db_connection.get_session() as session:
        # We minus 1 second just to make sure we ready that value
        gsp_yields: List[GSPYieldSQL] = get_gsp_yield(
            session=session,
            start_datetime_utc=start_utc_extra - timedelta(seconds=1),
            gsp_ids=list(range(1, N_GSPS + 1)),
        )

        logger.debug(f"Found {len(gsp_yields)} GSP yields from the database")

        gsp_yields_dict = []
        for gsp_yield in gsp_yields:
            location = Location.from_orm(gsp_yield.location)
            gsp_yield = GSPYield.from_orm(gsp_yield)

            gsp_yield_dict = gsp_yield.__dict__
            gsp_yield_dict["installed_capacity_mw"] = location.installed_capacity_mw
            gsp_yield_dict["solar_generation_mw"] = gsp_yield_dict["solar_generation_kw"] * 1000
            gsp_yield_dict["gsp_id"] = location.gsp_id
            gsp_yields_dict.append(gsp_yield_dict)

        gsp_yields_df = pd.DataFrame(gsp_yields_dict)

        logger.debug(gsp_yields_df.columns)

    if len(gsp_yields_df) == 0:
        logger.warning("Found no gsp yields, this might cause an error")
    else:
        logger.debug(f"Found {len(gsp_yields_df)} gsp yields")

    if len(gsp_yields_df) == 0:
        return pd.DataFrame(columns=["gsp_id"]), pd.DataFrame(columns=["gsp_id"])

    # pivot on
    gsp_yields_df = gsp_yields_df[
        ["datetime_utc", "gsp_id", "solar_generation_mw", "installed_capacity_mw"]
    ]
    logger.debug(gsp_yields_df.columns)
    gsp_yields_df.drop_duplicates(
        ["datetime_utc", "gsp_id", "solar_generation_mw"], keep="last", inplace=True
    )
    logger.debug(gsp_yields_df.columns)
    gsp_power_df = gsp_yields_df.pivot(
        index="datetime_utc", columns="gsp_id", values="solar_generation_mw"
    )

    gsp_capacity_df = gsp_yields_df.pivot(
        index="datetime_utc", columns="gsp_id", values="installed_capacity_mw"
    )

    logger.debug(f"{empty_df=}")
    logger.debug(f"{gsp_power_df=}")
    gsp_power_df = empty_df.join(gsp_power_df)
    gsp_capacity_df = empty_df.join(gsp_capacity_df)

    # interpolate in between, maximum 'live_interpolate_minutes' mins
    # note data is in 30 minutes chunks
    limit = int(interpolate_minutes / 30)
    if limit > 0:
        gsp_power_df.interpolate(
            limit=limit, inplace=True, method="cubic", fill_value="extrapolate"
        )
        gsp_capacity_df.interpolate(
            limit=limit, inplace=True, method="cubic", fill_value="extrapolate"
        )

    # filter out the extra minutes loaded
    logger.debug(f"{len(gsp_power_df)} of datetimes before filter on {start_utc}")
    gsp_power_df = gsp_power_df[gsp_power_df.index >= start_utc]
    gsp_capacity_df = gsp_capacity_df[gsp_capacity_df.index >= start_utc]
    logger.debug(f"{len(gsp_power_df)} of datetimes after filter on {start_utc}")

    # clip values to 0, this just stops any interpolation going below zero
    gsp_power_df.clip(lower=0, inplace=True)

    return gsp_power_df, gsp_capacity_df



"""  General Data Source Class """
import itertools
import logging
from concurrent import futures
from dataclasses import InitVar, dataclass
from numbers import Number
from pathlib import Path
from typing import Iterable, List, Tuple, Union

import pandas as pd
import xarray as xr

import nowcasting_dataset.filesystem.utils as nd_fs_utils
import nowcasting_dataset.time as nd_time
from nowcasting_dataset import square, utils
from nowcasting_dataset.data_sources.datasource_output import DataSourceOutput
from nowcasting_dataset.data_sources.metadata.metadata_model import SpaceTimeLocation
from nowcasting_dataset.dataset.xr_utils import (
    convert_coordinates_to_indexes_for_list_datasets,
    join_list_dataset_to_batch_dataset,
)
from nowcasting_dataset.utils import get_start_and_end_example_index

logger = logging.getLogger(__name__)


@dataclass
class DataSource:
    """Abstract base class.

    Attributes:
      history_minutes: Number of minutes of history to include in each example.
        Does NOT include t0.  That is, if history_minutes = 0 then the example
        will start at t0.
      forecast_minutes: Number of minutes of forecast to include in each example.
        Does NOT include t0.  If forecast_minutes = 0 then the example will end
        at t0.  If both history_minutes and forecast_minutes are 0, then the example
        will consist of a single timestep at t0.
      sample_period_minutes: The time delta between each data point.  Note that this is set
        using the sample_period_minutes property, so it can be overridden by child classes.

    Attributes ending in `_length` are sequence lengths represented
        as integer numbers of timesteps.
    Attributes ending in `_duration` are sequence durations represented as pd.Timedeltas.
    """

    history_minutes: int
    forecast_minutes: int

    def __post_init__(self):
        """Post Init"""
        self.check_input_paths_exist()
        self.sample_period_duration = pd.Timedelta(self.sample_period_minutes, unit="minutes")

        # TODO: Do we still need all these different representations of sequence lengths?
        # See GitHub issue #219 for more details, and to track this TODO task.
        self.history_length = self.history_minutes // self.sample_period_minutes
        self.forecast_length = self.forecast_minutes // self.sample_period_minutes

        assert self.history_length >= 0
        assert self.forecast_length >= 0
        assert self.history_minutes % self.sample_period_minutes == 0, (
            f"sample period ({self.sample_period_minutes}) minutes "
            f"does not fit into historic minutes ({self.history_minutes})"
        )
        assert self.forecast_minutes % self.sample_period_minutes == 0, (
            f"sample period ({self.sample_period_minutes}) minutes "
            f"does not fit into forecast minutes ({self.forecast_minutes})"
        )

        # Plus 1 because neither history_length nor forecast_length include t0.
        self.total_seq_length = self.history_length + self.forecast_length + 1

        self.history_duration = pd.Timedelta(self.history_minutes, unit="minutes")
        self.forecast_duration = pd.Timedelta(self.forecast_minutes, unit="minutes")
        # Add sample_period_duration because neither history_duration not forecast_duration
        # include t0.
        self.total_seq_duration = (
            self.history_duration + self.forecast_duration + self.sample_period_duration
        )

    def _get_start_dt(
        self, t0_datetime_utc: Union[pd.Timestamp, pd.DatetimeIndex]
    ) -> Union[pd.Timestamp, pd.DatetimeIndex]:

        return t0_datetime_utc - self.history_duration

    def _get_end_dt(
        self, t0_datetime_utc: Union[pd.Timestamp, pd.DatetimeIndex]
    ) -> Union[pd.Timestamp, pd.DatetimeIndex]:
        return t0_datetime_utc + self.forecast_duration

    def get_contiguous_t0_time_periods(self) -> pd.DataFrame:
        """Get all time periods which contain valid t0 datetimes.

        `t0` is the datetime of the most recent observation.

        Returns:
          pd.DataFrame where each row represents a single time period.  The pd.DataFrame
          has two columns: `start_dt` and `end_dt` (where 'dt' is short for 'datetime').

        Raises:
          NotImplementedError if this DataSource has no concept of a datetime index.
        """
        contiguous_time_periods = self.get_contiguous_time_periods()
        contiguous_time_periods["start_dt"] += self.history_duration
        contiguous_time_periods["end_dt"] -= self.forecast_duration
        assert (contiguous_time_periods["start_dt"] <= contiguous_time_periods["end_dt"]).all()
        return contiguous_time_periods

    # ************* METHODS THAT CAN BE OVERRIDDEN ****************************
    @property
    def sample_period_minutes(self) -> int:
        """
        This is the default sample period in minutes.

        This functions may be overwritten if the sample period of the data source is not 5 minutes.
        """
        logger.debug(
            "Getting sample_period_minutes default of 5 minutes. "
            "This means the data is spaced 5 minutes apart"
        )
        return 5

    def open(self):
        """Open the data source, if necessary.

        Called from each worker process.  Useful for data sources where the
        underlying data source cannot be forked (like Zarr).

        Data sources which can be forked safely should call open() from __init__().
        """
        pass

    # TODO: Issue #319: Standardise parameter names.
    # TODO: Issue #367: Reduce duplication.
    @utils.exception_logger
    def create_batches(
        self,
        spatial_and_temporal_locations_of_each_example: List[SpaceTimeLocation],
        idx_of_first_batch: int,
        batch_size: int,
        dst_path: Path,
        local_temp_path: Path,
        upload_every_n_batches: int,
    ) -> None:
        """Create multiple batches and save them to disk.

        Safe to call from worker processes.

        Args:
            spatial_and_temporal_locations_of_each_example (pd.DataFrame): A DataFrame where each
                row specifies the spatial and temporal location of an example. The number of rows
                must be an exact multiple of `batch_size`.
                Columns are: t0_datetime_UTC, x_center_OSGB, y_center_OSGB.
            idx_of_first_batch (int): The batch number of the first batch to create.
            batch_size (int): The number of examples per batch.
            dst_path (Path): The final destination path for the batches.  Must exist.
            local_temp_path (Path): The local temporary path.  This is only required when dst_path
                is a cloud storage bucket, so files must first be created on the VM's local disk in
                temp_path and then uploaded to dst_path every `upload_every_n_batches`. Must exist.
                Will be emptied.
            upload_every_n_batches (int): Upload the contents of temp_path to dst_path after this
                number of batches have been created.  If 0 then will write directly to `dst_path`.
        """
        # Sanity checks:
        assert (
            idx_of_first_batch >= 0
        ), "The batch number of the first batch to create should be greater than 0"
        assert batch_size > 0, "The batch size should be strictly greater than 0."
        assert len(spatial_and_temporal_locations_of_each_example) % batch_size == 0, (
            f"{len(spatial_and_temporal_locations_of_each_example)=} must be"
            f" exactly divisible by {batch_size=}"
        )
        assert upload_every_n_batches >= 0, "`upload_every_n_batches` must be >= 0"

        self.open()

        # Figure out where to write batches to:
        save_batches_locally_and_upload = upload_every_n_batches > 0
        if save_batches_locally_and_upload:
            nd_fs_utils.delete_all_files_in_temp_path(local_temp_path)
        path_to_write_to = local_temp_path if save_batches_locally_and_upload else dst_path

        # Split locations per example into batches:
        n_batches = len(spatial_and_temporal_locations_of_each_example) // batch_size
        locations_for_batches = []
        for batch_idx in range(n_batches):
            start_example_idx, end_example_idx = get_start_and_end_example_index(
                batch_idx=batch_idx, batch_size=batch_size
            )

            locations_for_batch = spatial_and_temporal_locations_of_each_example[
                start_example_idx:end_example_idx
            ]
            locations_for_batches.append(locations_for_batch)

        # Loop round each batch:
        for n_batches_processed, locations_for_batch in enumerate(locations_for_batches):
            batch_idx = idx_of_first_batch + n_batches_processed
            logger.debug(f"{self.__class__.__name__} creating batch {batch_idx}!")

            # Generate batch.
            batch = self.get_batch(locations=locations_for_batch)

            # Save batch to disk.
            batch.save_netcdf(
                batch_i=batch_idx, path=path_to_write_to, add_data_source_name_to_path=False
            )

            # Upload if necessary.
            if (
                save_batches_locally_and_upload
                and n_batches_processed > 0
                and n_batches_processed % upload_every_n_batches == 0
            ):
                nd_fs_utils.upload_and_delete_local_files(
                    dst_path=dst_path, local_path=path_to_write_to
                )

        # Upload last few batches, if necessary:

        if save_batches_locally_and_upload:
            nd_fs_utils.upload_and_delete_local_files(
                dst_path=dst_path, local_path=path_to_write_to
            )

    def get_batch(self, locations: List[SpaceTimeLocation]) -> DataSourceOutput:
        """
        Get Batch Data

        Args:
            locations: List of locations object
                A location object contains
                - a timestamp of the example (t0_datetime_utc),
                - the x center location of the example (x_location_osgb)
                - the y center location of the example(y_location_osgb)

        Returns: Batch data.
        """

        batch_size = len(locations)

        with futures.ThreadPoolExecutor(max_workers=batch_size) as executor:
            future_examples = []
            for location in locations:
                future_example = executor.submit(self.get_example, location)
                future_examples.append(future_example)

            # Get the examples back.  Loop round each future so we can log a helpful error.
            # If the worker thread raised an exception then the exception won't "bubble up"
            # until we call future_example.result().
            examples = []
            for example_i, future_example in enumerate(future_examples):
                try:
                    result = future_example.result()
                except Exception:
                    logger.error(f"Exception when processing {example_i=}!")
                    raise
                else:
                    examples.append(result)

        # Get the DataSource class, this could be one of the data sources like Sun
        cls = self.get_data_model_for_batch()

        # Set the coords to be indices before joining into a batch
        examples = convert_coordinates_to_indexes_for_list_datasets(examples)

        # join the examples together, and cast them to the cls, so that validation can occur
        batch_one_datasource = cls(join_list_dataset_to_batch_dataset(examples))

        # lets validate
        cls.validate(batch_one_datasource)

        return batch_one_datasource

    def datetime_index(self) -> pd.DatetimeIndex:
        """Returns a complete list of all available datetimes."""
        # Leave this NotImplemented if this DataSource has no concept
        # of a list of datetimes (e.g. for DatetimeDataSource).
        raise NotImplementedError(f"Datetime not implemented for class {self.__class__}")

    @staticmethod
    def get_data_model_for_batch():
        """Get the model that is used in the batch"""
        raise NotImplementedError()

    def get_contiguous_time_periods(self) -> pd.DataFrame:
        """Get all the time periods for which this DataSource has contiguous data.

        Optionally filter out any time periods which don't make sense for this DataSource,
        e.g. remove nighttime.

        Returns:
          pd.DataFrame where each row represents a single time period.  The pd.DataFrame
          has two columns: `start_dt` and `end_dt` (where 'dt' is short for 'datetime').

        Raises:
          NotImplementedError if this DataSource has no concept of a datetime index.
        """
        datetimes = self.datetime_index()
        return nd_time.get_contiguous_time_periods(
            datetimes=datetimes,
            min_seq_length=self.total_seq_length,
            max_gap_duration=self.sample_period_duration,
        )

    def get_locations(
        self, t0_datetimes_utc: pd.DatetimeIndex
    ) -> Tuple[List[Number], List[Number]]:
        """Find a valid geographical locations for each t0_datetime.

        Should be overridden by DataSources which may be used to define the locations.

        Returns:  x_locations, y_locations. Each has one entry per t0_datetime.
            Locations are in OSGB coordinates.
        """
        raise NotImplementedError()

    def get_all_locations(
        self, t0_datetimes_utc: pd.DatetimeIndex
    ) -> Tuple[pd.DatetimeIndex, List[Number], List[Number]]:
        """Find all valid geographical locations for each t0_datetime.

        Should be overridden by DataSources which may be used to define the locations.

        Returns:  all_t0_datetimes_utc, x_locations, y_locations.
            For each single t0_datetime, there are all possible locations.
            Each return has length len(t0_datetimes_utc) * number_of_location
            Locations are in OSGB coordinates.
        """
        raise NotImplementedError()

    def get_number_locations(self) -> int:
        """
        Get the number of locations of a data source

        For GSP, this is the number of GSP
        For PV, this is the number of PV locations
        """
        raise NotImplementedError()

    # ****************** METHODS THAT MUST BE OVERRIDDEN **********************
    def _get_time_slice(self, t0_datetime_utc: pd.Timestamp):
        """Get a single timestep of data.  Must be overridden if get_example is not overridden."""
        raise NotImplementedError()

    def get_example(
        self,
        location: SpaceTimeLocation,  #: Location object of the most recent observation
    ) -> xr.Dataset:
        """Must be overridden by child classes."""
        raise NotImplementedError()

    def check_input_paths_exist(self) -> None:
        """Check any input paths exist.  Raise FileNotFoundError if not.

        Must be overridden by child classes.
        """
        raise NotImplementedError()


@dataclass
class ImageDataSource(DataSource):
    """Abstract base class for image Data source."""

    image_size_pixels_height: InitVar[int]
    image_size_pixels_width: InitVar[int]
    meters_per_pixel: InitVar[int]

    def __post_init__(
        self, image_size_pixels_height: int, image_size_pixels_width: int, meters_per_pixel: int
    ):
        """Post Init"""
        super().__post_init__()
        self._rectangle = square.Rectangle(
            size_pixels_height=image_size_pixels_height,
            size_pixels_width=image_size_pixels_width,
            meters_per_pixel=meters_per_pixel,
        )


@dataclass
class ZarrDataSource(ImageDataSource):
    """
    A General Zarr Data source

    Attributes:
      _data: xr.DataArray data, opened by open().
        x is left-to-right.
        y is top-to-bottom.
        Access using public data property.
      consolidated: Whether or not the Zarr store is consolidated.
      channels: The Zarr parameters to load.
    """

    # zarr_path and channels must be set.  But dataclasses complains about defining a non-default
    # argument after a default argument if we remove the ` = None`.
    zarr_path: Union[Path, str] = None
    channels: Iterable[str] = None
    consolidated: bool = True

    def __post_init__(
        self, image_size_pixels_height: int, image_size_pixels_width: int, meters_per_pixel: int
    ):
        """Post init"""
        super().__post_init__(image_size_pixels_height, image_size_pixels_width, meters_per_pixel)
        self._data = None

    def check_input_paths_exist(self) -> None:
        """Check input paths exist.  If not, raise a FileNotFoundError."""
        nd_fs_utils.check_path_exists(self.zarr_path)

    @property
    def data(self):
        """Data property"""
        if self._data is None:
            raise RuntimeError("Please run `open()` before accessing data!")
        return self._data

    def get_example(self, location: SpaceTimeLocation) -> xr.Dataset:
        """
        Get Example data

        Args:
            location: A location object of the example which contains
                - a timestamp of the example (t0_datetime_utc),
                - the x center location of the example (x_location_osgb)
                - the y center location of the example(y_location_osgb)

        Returns: Example Data

        """

        t0_datetime_utc = location.t0_datetime_utc
        x_center_osgb = location.x_center_osgb
        y_center_osgb = location.y_center_osgb

        logger.debug(
            f"Getting example for {t0_datetime_utc=},  " f"{x_center_osgb=} and  {y_center_osgb=}"
        )

        selected_data = self._get_time_slice(t0_datetime_utc)
        bounding_box = self._rectangle.bounding_box_centered_on(
            x_center_osgb=x_center_osgb, y_center_osgb=y_center_osgb
        )
        selected_data = selected_data.sel(
            x_osgb=slice(bounding_box.left, bounding_box.right),
            y_osgb=slice(bounding_box.top, bounding_box.bottom),
        )

        # selected_sat_data is likely to have 1 too many pixels in x and y
        # because sel(x=slice(a, b)) is [a, b], not [a, b).  So trim:
        selected_data = selected_data.isel(
            x_osgb=slice(0, self._rectangle.size_pixels_width),
            y_osgb=slice(0, self._rectangle.size_pixels_height),
        )

        selected_data = self._post_process_example(selected_data, t0_datetime_utc)

        if selected_data.shape != self._shape_of_example:
            m = (
                "Example is wrong shape! "
                f"x_center_osgb={x_center_osgb}\n"
                f"y_center_osgb={y_center_osgb}\n"
                f"t0_datetime_utc={t0_datetime_utc}\n"
                f"times are {selected_data.time}\n"
                f"expected shape={self._shape_of_example}\n"
                f"actual shape {selected_data.shape}"
            )
            logger.error(m)
            raise RuntimeError(m)

        return selected_data.load().to_dataset(name="data")

    def geospatial_border(self) -> list[tuple[Number, Number]]:
        """
        Get 'corner' coordinates for a rectangle within the boundary of the data.

        Returns List of 2-tuples of the x and y coordinates of each corner,
        in OSGB projection.
        """
        GEO_BORDER: int = 64  #: In same geo projection and units as sat_data.
        data = self._open_data()
        return [
            (data.x.values[x], data.y.values[y])
            for x, y in itertools.product([GEO_BORDER, -GEO_BORDER], [GEO_BORDER, -GEO_BORDER])
        ]

    # ****************** METHODS THAT CAN BE OVERRIDDEN **********************
    def _post_process_example(
        self, selected_data: xr.DataArray, t0_dt: pd.Timestamp
    ) -> xr.DataArray:
        return selected_data

    # ****************** METHODS THAT MUST BE OVERRIDDEN **********************
    # (in addition to the DataSource methods that must be overridden)
    def open(self) -> None:
        """
        Open the data

        We don't want to _open_data() in __init__.
        If we did that, then we couldn't copy ZarrDataSource
        instances into separate processes.  Instead,
        call open() _after_ creating separate processes.
        """
        raise NotImplementedError()

    def _open_data(self) -> xr.DataArray:
        raise NotImplementedError()



""" Util functions for PV data source"""
from typing import List

from nowcasting_dataset.consts import PV_PROVIDERS


def encode_label(indexes: List[str], label: str) -> List[str]:
    """
    Encode the label to a list of indexes.

    The new encoding must be integers and unique.
    It would be useful if the indexes can read and deciphered by humans.
    This is done by times the original index by 10
    and adding 1 for passiv or 2 for other lables

    Args:
        indexes: list of indexes
        label: either 'passiv' or 'pvoutput'

    Returns: list of indexes encoded by label
    """
    assert label in PV_PROVIDERS
    # this encoding does work if the number of pv providers is more than 10
    assert len(PV_PROVIDERS) < 10

    label_index = PV_PROVIDERS.index(label)
    new_index = [str(int(col) * 10 + label_index) for col in indexes]

    return new_index



"""Base Manager class."""

import logging
from pathlib import Path
from typing import Optional

import nowcasting_dataset.utils as nd_utils
from nowcasting_dataset import config
from nowcasting_dataset.data_sources import ALL_DATA_SOURCE_NAMES, MAP_DATA_SOURCE_NAME_TO_CLASS

logger = logging.getLogger(__name__)


class ManagerBase:
    """The Manager initializes and manage a dict of DataSource objects.

    Attrs:
      config: Configuration object.
      data_sources: dict[str, DataSource]
      data_source_which_defines_geospatial_locations: DataSource: The DataSource used to compute the
        geospatial locations of each example.
      save_batches_locally_and_upload: bool: Set to True by `load_yaml_configuration()` if
        `config.process.upload_every_n_batches > 0`.
    """

    def __init__(self) -> None:  # noqa: D107
        self.config = None
        self.data_sources = {}
        self.data_source_which_defines_geospatial_locations = None

    def load_yaml_configuration(self, filename: str, set_git: bool = True) -> None:
        """Load YAML config from `filename`."""
        logger.debug(f"Loading YAML configuration file {filename}")
        self.config = config.load_yaml_configuration(filename)
        if set_git:
            self.config = config.set_git_commit(self.config)
        self.save_batches_locally_and_upload = self.config.process.upload_every_n_batches > 0
        logger.debug(f"config={self.config}")

    def initialize_data_sources(
        self, names_of_selected_data_sources: Optional[list[str]] = ALL_DATA_SOURCE_NAMES
    ) -> None:
        """Initialize DataSources specified in the InputData configuration.

        For each key in each DataSource's configuration object, the string `<data_source_name>_`
        is removed from the key before passing to the DataSource constructor.  This allows us to
        have verbose field names in the configuration YAML files, whilst also using standard
        constructor arguments for DataSources.
        """
        for data_source_name in names_of_selected_data_sources:
            logger.debug(f"Creating {data_source_name} DataSource object.")
            config_for_data_source = getattr(self.config.input_data, data_source_name)
            if config_for_data_source is None:
                logger.info(f"No configuration found for {data_source_name}.")
                continue
            config_for_data_source = config_for_data_source.dict()
            config_for_data_source.pop("log_level")

            # save config to data source logger
            data_source_logger = logging.getLogger(
                f"nowcasting_dataset.data_sources.{data_source_name}"
            )
            data_source_logger.debug(
                f"The configuration for {data_source_name} is {config_for_data_source}"
            )

            # Strip `<data_source_name>_` from the config option field names.
            config_for_data_source = nd_utils.remove_regex_pattern_from_keys(
                config_for_data_source, pattern_to_remove=f"^{data_source_name}_"
            )

            # TODO: #631 remove
            if data_source_name == "pv":
                config_for_data_source.pop("filename")
                config_for_data_source.pop("metadata_filename")

            data_source_class = MAP_DATA_SOURCE_NAME_TO_CLASS[data_source_name]
            try:
                data_source = data_source_class(**config_for_data_source)
            except Exception:
                logger.exception(
                    f"Exception whilst instantiating {data_source_name}! "
                    f"Tried with configuration {config_for_data_source} "
                    f"in {data_source_class}"
                )
                raise
            self.data_sources[data_source_name] = data_source

        # Set data_source_which_defines_geospatial_locations:
        try:
            self.data_source_which_defines_geospatial_locations = self.data_sources[
                self.config.input_data.data_source_which_defines_geospatial_locations
            ]
        except KeyError:
            if self._locations_csv_file_exists():
                logger.info(
                    f"{self.config.input_data.data_source_which_defines_geospatial_locations=}"
                    " is not a member of the DataSources, but that does not matter because the CSV"
                    " files which specify the locations of the examples already exists!"
                )
            else:
                msg = (
                    "input_data.data_source_which_defines_geospatial_locations="
                    f"{self.config.input_data.data_source_which_defines_geospatial_locations}"
                    " is not a member of the DataSources, so cannot set"
                    " self.data_source_which_defines_geospatial_locations!"
                    f" The available DataSources are: {list(self.data_sources.keys())}"
                )
                logger.error(msg)
                raise RuntimeError(msg)
        else:
            logger.info(
                f"DataSource `{data_source_name}` set as"
                " data_source_which_defines_geospatial_locations."
            )

    def _locations_csv_file_exists(self):
        return False

    def _filename_of_locations_csv_file(self, split_name: str) -> Path:
        return self.config.output_data.filepath / split_name



""" Function to check Square"""
import pandas as pd

from nowcasting_dataset.square import BoundingBox, Square, get_closest_coordinate_order


def test_init():
    """Set up square"""
    square = Square(size_pixels=128, meters_per_pixel=1000)
    assert square._half_size_meters == 64_000


def test_bounding_box_centered_on():
    """Get bounding box"""
    square = Square(size_pixels=50, meters_per_pixel=1000)
    bounding_box = square.bounding_box_centered_on(x_center_osgb=20_000, y_center_osgb=40_000)
    assert bounding_box == BoundingBox(top=65_000, bottom=15_000, left=-5_000, right=45_000)


def test_get_closest_coordinate_order():
    """Test get_closest_coordinate_order"""
    x_center = 5
    y_center = 5

    true_order = [1, 0, 2]
    x = pd.Series(data=[4, 5, 10])
    y = pd.Series(data=[6, 5, 0])

    order = get_closest_coordinate_order(x_center=x_center, y_center=y_center, x=x, y=y)

    assert list(order.values) == true_order



""" Time functions """
import logging
import random
import warnings
from datetime import timedelta
from typing import Dict, Iterable, List, Optional, Tuple

import numpy as np
import pandas as pd
import pvlib

from nowcasting_dataset import geospatial, utils

logger = logging.getLogger(__name__)


FIVE_MINUTES = pd.Timedelta("5 minutes")
THIRTY_MINUTES = pd.Timedelta("30 minutes")
ONE_HOUR = pd.Timedelta("1 hour")


def select_daylight_datetimes(
    datetimes: pd.DatetimeIndex,
    locations: Iterable[Tuple[float, float]],
    ghi_threshold: float = 10,
    keep_dawn_dusk_hours: int = 0,
) -> pd.DatetimeIndex:
    """
    Select only the day time datetimes

    Args:
        datetimes: DatetimeIndex to filter.
        locations: List of Tuples of x, y coordinates in OSGB projection.
        For example, use the four corners of the satellite imagery.
        ghi_threshold: Global horizontal irradiance threshold.
          (Watts per square meter?)
        keep_dawn_dusk_hours: option to keep extra hours before dawn, or after dusk.
            This is defaulted to zero

    Returns: datetimes for which the global horizontal irradiance (GHI) is above ghi_threshold
    across all locations.

    """

    if keep_dawn_dusk_hours > 0:
        # get dawn datetimes by
        # - adding X hours the original datetimes,
        # - running through select_daylight_datetimes
        # - minusing X hours from the results
        # similar for dusk
        temp_keep_dawn_dusk_hours = 0
        dawn = select_daylight_datetimes(
            datetimes=datetimes + timedelta(hours=keep_dawn_dusk_hours),
            locations=locations,
            ghi_threshold=ghi_threshold,
            keep_dawn_dusk_hours=temp_keep_dawn_dusk_hours,
        ) - timedelta(hours=keep_dawn_dusk_hours)

        dusk = select_daylight_datetimes(
            datetimes=datetimes - timedelta(hours=keep_dawn_dusk_hours),
            locations=locations,
            ghi_threshold=ghi_threshold,
            keep_dawn_dusk_hours=temp_keep_dawn_dusk_hours,
        ) + timedelta(hours=keep_dawn_dusk_hours)

        # take union of dusk and dawn
        return dusk.union(dawn)

    ghi_for_all_locations = []
    for x, y in locations:
        lat, lon = geospatial.osgb_to_lat_lon(x, y)
        location = pvlib.location.Location(latitude=lat, longitude=lon)
        with warnings.catch_warnings():
            # PyTables triggers a DeprecationWarning in Numpy >= 1.20:
            # "tables/array.py:241: DeprecationWarning: `np.object` is a
            # deprecated alias for the builtin `object`."
            # See https://github.com/PyTables/PyTables/issues/898
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            clearsky = location.get_clearsky(datetimes)
        ghi = clearsky["ghi"]
        ghi_for_all_locations.append(ghi)

    ghi_for_all_locations = pd.concat(ghi_for_all_locations, axis="columns")
    max_ghi = ghi_for_all_locations.max(axis="columns")
    mask = max_ghi > ghi_threshold
    return datetimes[mask]


def single_period_to_datetime_index(period: pd.Series, freq: str) -> pd.DatetimeIndex:
    """Return a DatetimeIndex from period['start_dt'] to period['end_dt'] at frequency freq.

    Before computing the date_range, this function first takes the ceiling of the
    start_dt (at frequency `freq`); and takes the floor of end_dt.  For example,
    if `freq` is '5 minutes' and `start_dt` is 12:03, then the ceiling of `start_dt`
    will be 12:05.  This is done so that all the returned datetimes are aligned to `freq`
    (e.g. if `freq` is '5 minutes' then every returned datetime will be at 00, 05, ..., 55
    minutes past the hour.
    """
    start_dt = period["start_dt"].ceil(freq)
    end_dt = period["end_dt"].floor(freq)
    return pd.date_range(start_dt, end_dt, freq=freq)


def time_periods_to_datetime_index(time_periods: pd.DataFrame, freq: str) -> pd.DatetimeIndex:
    """Convert a DataFrame of time periods into a DatetimeIndex at a particular frequency.

    See the docstring of intersection_of_2_dataframes_of_periods() for more details.
    """
    assert len(time_periods) > 0
    dt_index = single_period_to_datetime_index(time_periods.iloc[0], freq=freq)
    for _, time_period in time_periods.iloc[1:].iterrows():
        new_dt_index = single_period_to_datetime_index(time_period, freq=freq)
        dt_index = dt_index.union(new_dt_index)
    return dt_index


def intersection_of_multiple_dataframes_of_periods(
    time_periods: list[pd.DataFrame],
) -> pd.DataFrame:
    """Finds the intersection of a list of time periods.

    See the docstring of intersection_of_2_dataframes_of_periods() for more details.
    """
    assert len(time_periods) > 0
    if len(time_periods) == 1:
        return time_periods[0]
    intersection = intersection_of_2_dataframes_of_periods(time_periods[0], time_periods[1])
    for time_period in time_periods[2:]:
        intersection = intersection_of_2_dataframes_of_periods(intersection, time_period)
    return intersection


def intersection_of_2_dataframes_of_periods(a: pd.DataFrame, b: pd.DataFrame) -> pd.DataFrame:
    """Finds the intersection of two pd.DataFrames of time periods.

    Each row of each pd.DataFrame represents a single time period.  Each pd.DataFrame has
    two columns: `start_dt` and `end_dt` (where 'dt' is short for 'datetime').

    A typical use-case is that each pd.DataFrame represents all the time periods where
    a `DataSource` has contiguous, valid data.

    Here's a graphical example of two pd.DataFrames of time periods and their intersection:

                 ----------------------> TIME ->---------------------
               a: |-----|   |----|     |----------|     |-----------|
               b:    |--------|                       |----|    |---|
    intersection:    |--|   |-|                         |--|    |---|

    Args:
        a, b: pd.DataFrame where each row represents a time period.  The pd.DataFrame has
        two columns: start_dt and end_dt.

    Returns:
        Sorted list of intersecting time periods represented as a pd.DataFrame with two columns:
        start_dt and end_dt.
    """
    if a.empty or b.empty:
        return pd.DataFrame(columns=["start_dt", "end_dt"])

    all_intersecting_periods = []
    for a_period in a.itertuples():
        # Five ways in which two periods may overlap:
        # a: |----| or |---|   or  |---| or   |--|   or |-|
        # b:  |--|       |---|   |---|      |------|    |-|
        # In all five, `a` must always start before `b` ends,
        # and `a` must always end after `b` starts:
        overlapping_periods = b[(a_period.start_dt < b.end_dt) & (a_period.end_dt > b.start_dt)]

        # There are two ways in which two periods may *not* overlap:
        # a: |---|        or        |---|
        # b:       |---|      |---|
        # `overlapping` will not include periods which do *not* overlap.

        # Now find the intersection of each period in `overlapping_periods` with
        # the period from `a` that starts at `a_start_dt` and ends at `a_end_dt`.
        # We do this by clipping each row of `overlapping_periods`
        # to start no earlier than `a_start_dt`, and end no later than `a_end_dt`.

        # First, make a copy, so we don't clip the underlying data in `b`.
        intersecting_periods = overlapping_periods.copy()
        intersecting_periods.start_dt.clip(lower=a_period.start_dt, inplace=True)
        intersecting_periods.end_dt.clip(upper=a_period.end_dt, inplace=True)

        all_intersecting_periods.append(intersecting_periods)

    all_intersecting_periods = pd.concat(all_intersecting_periods)
    return all_intersecting_periods.sort_values(by="start_dt").reset_index(drop=True)


def get_contiguous_time_periods(
    datetimes: pd.DatetimeIndex,
    min_seq_length: int,
    max_gap_duration: pd.Timedelta = THIRTY_MINUTES,
) -> pd.DataFrame:
    """Return a pd.DataFrame where each row records the boundary of a contiguous time period.

    Args:
      datetimes: The pd.DatetimeIndex of the timeseries. Must be sorted.
      min_seq_length: Sequences of min_seq_length or shorter will be discarded.  Typically, this
        would be set to the `total_seq_length` of each machine learning example.
      max_gap_duration: If any pair of consecutive `datetimes` is more than `max_gap_duration`
        apart, then this pair of `datetimes` will be considered a "gap" between two contiguous
        sequences. Typically, `max_gap_duration` would be set to the sample period of
        the timeseries.

    Returns:
      pd.DataFrame where each row represents a single time period.  The pd.DataFrame
          has two columns: `start_dt` and `end_dt` (where 'dt' is short for 'datetime').
    """
    # Sanity checks.
    assert len(datetimes) > 0
    assert min_seq_length > 1
    assert datetimes.is_monotonic_increasing
    assert datetimes.is_unique

    # Find indices of gaps larger than max_gap:
    gap_mask = np.diff(datetimes) > max_gap_duration
    gap_indices = np.argwhere(gap_mask)[:, 0]

    # gap_indicies are the indices into dt_index for the timestep immediately
    # *before* the gap.  e.g. if the datetimes at 12:00, 12:05, 18:00, 18:05
    # then gap_indicies will be [1]. So we add 1 to gap_indices to get
    # segment_boundaries, an index into dt_index which identifies the _start_
    # of each segment.
    segment_boundaries = gap_indices + 1

    # Capture the last segment of dt_index.
    segment_boundaries = np.concatenate((segment_boundaries, [len(datetimes)]))

    periods: List[Dict[str, pd.Timestamp]] = []
    start_i = 0
    for next_start_i in segment_boundaries:
        n_timesteps = next_start_i - start_i
        if n_timesteps > min_seq_length:
            end_i = next_start_i - 1
            period = {"start_dt": datetimes[start_i], "end_dt": datetimes[end_i]}
            periods.append(period)
        start_i = next_start_i

    assert len(periods) > 0

    return pd.DataFrame(periods)


def make_random_time_vectors(
    batch_size,
    seq_length_5_minutes,
    seq_length_30_minutes,
    seq_length_60_minutes: Optional[int] = None,
) -> dict:
    """
    Make random time vectors

    1. t0_dt, Get random datetimes from 2019
    2. Exapnd t0_dt to make 5 and 30 mins sequences

    Args:
        batch_size: the batch size
        seq_length_5_minutes: the length of the sequence in 5 mins deltas
        seq_length_30_minutes: the length of the sequence in 30 mins deltas
        seq_length_60_minutes: the length of the sequence in 60 mins deltas

    Returns:
        - t0_dt: [batch_size] random init datetimes
        - time_5: [batch_size, seq_length_5_minutes] random sequence of datetimes, with
            5 mins deltas. t0_dt is in the middle of the sequence
        - time_30: [batch_size, seq_length_30_minutes] random sequence of datetimes, with
            30 mins deltas. t0_dt is in the middle of the sequence
        - time_60 : [batch_size, seq_length_60_minutes] random sequence of datetimes, with
            60 mins deltas. t0_dt is in the middle of the sequence
    """
    delta_5 = pd.Timedelta(minutes=5)
    delta_30 = pd.Timedelta(minutes=30)

    if seq_length_60_minutes is None:
        seq_length_60_minutes = int(seq_length_5_minutes / 12)

    data_range = pd.date_range("2019-01-01", "2021-01-01", freq="5T")
    t0_dt = pd.Series(random.choices(data_range, k=batch_size))
    time_5 = (
        pd.DataFrame([t0_dt + i * delta_5 for i in range(seq_length_5_minutes)])
        - int(seq_length_5_minutes / 2) * delta_5
    )
    time_30 = (
        pd.DataFrame([t0_dt + i * delta_30 for i in range(seq_length_30_minutes)])
        - int(seq_length_30_minutes / 2) * delta_5
    )
    time_60 = (
        pd.DataFrame([t0_dt + i * ONE_HOUR for i in range(seq_length_60_minutes)])
        - int(seq_length_60_minutes / 2) * delta_5
    )

    return {
        "t0_dt": utils.to_numpy(t0_dt).astype(np.int32),
        "time_5": utils.to_numpy(time_5.T).astype(np.int32),
        "time_30": utils.to_numpy(time_30.T).astype(np.int32),
        "time_60": utils.to_numpy(time_60.T).astype(np.int32),
    }


def floor_minutes_dt(dt, minutes: Optional[int] = 30):
    """
    Floor a datetime by 30 mins.

    For example:
    2021-01-01 17:01:01 --> 2021-01-01 17:00:00
    2021-01-01 17:35:01 --> 2021-01-01 17:30:00

    :param dt:
    :return:
    """
    approx = np.floor(dt.minute / float(minutes)) * minutes
    dt = dt.replace(minute=0)
    dt = dt.replace(second=0)
    dt = dt.replace(microsecond=0)
    dt += timedelta(minutes=approx)

    return dt



""" NWP Data Source """
import io
import logging
from dataclasses import InitVar, dataclass
from typing import Iterable, Optional

import fsspec
import numpy as np
import pandas as pd
import xarray as xr

from nowcasting_dataset import utils
from nowcasting_dataset.consts import NWP_VARIABLE_NAMES
from nowcasting_dataset.data_sources.data_source import ZarrDataSource
from nowcasting_dataset.data_sources.nwp.nwp_model import NWP

_LOG = logging.getLogger(__name__)


@dataclass
class NWPDataSource(ZarrDataSource):
    """
    NWP Data Source (Numerical Weather Predictions)

    Attributes:
        _data: xr.DataArray of Numerical Weather Predictions, opened by open().
            x is left-to-right.
            y is top-to-bottom (after reversing the `y` index in open_nwp()).
        consolidated: Whether or not the Zarr store is consolidated.
        channels: The NWP forecast parameters to load. If None then don't filter.
            See:  http://cedadocs.ceda.ac.uk/1334/1/uk_model_data_sheet_lores1.pdf
            All of these params are "instant" (i.e. a snapshot at the target time,
            not accumulated over some time period).  The available params are:
                cdcb  : Height of lowest cloud base > 3 oktas, in meters above surface.
                lcc   : Low-level cloud cover in %.
                mcc   : Medium-level cloud cover in %.
                hcc   : High-level cloud cover in %.
                sde   : Snow depth in meters.
                hcct  : Height of convective cloud top, meters above surface.
                        WARNING: hcct has NaNs where there are no clouds forecast to exist!
                dswrf : Downward short-wave radiation flux in W/m^2 (irradiance) at surface.
                dlwrf : Downward long-wave radiation flux in W/m^2 (irradiance) at surface.
                h     : Geometrical height, meters.
                t     : Air temperature at 1 meter above surface in Kelvin.
                r     : Relative humidty in %.
                dpt   : Dew point temperature in Kelvin.
                vis   : Visibility in meters.
                si10  : Wind speed in meters per second, 10 meters above surface.
                wdir10: Wind direction in degrees, 10 meters above surface.
                prmsl : Pressure reduce to mean sea level in Pascals.
                prate : Precipitation rate at the surface in kg/m^2/s.
    """

    channels: Optional[Iterable[str]] = NWP_VARIABLE_NAMES
    image_size_pixels_height: InitVar[int] = 2
    image_size_pixels_width: InitVar[int] = 2
    meters_per_pixel: InitVar[int] = 2_000

    def __post_init__(
        self, image_size_pixels_height: int, image_size_pixels_width: int, meters_per_pixel: int
    ):
        """
        Post init

        Args:
            image_size_pixels_height: number of pixels in image height
            image_size_pixels_width: number of pixels in image width
            meters_per_pixel: how many meteres for each pixel

        """
        super().__post_init__(image_size_pixels_height, image_size_pixels_width, meters_per_pixel)
        n_channels = len(self.channels)
        self._shape_of_example = (
            n_channels,
            self.total_seq_length,
            image_size_pixels_height,
            image_size_pixels_width,
        )
        _LOG.info(
            f"NWPDataSource instantiated with {self.channels=},"
            f" {self._rectangle.size_pixels_height=},"
            f" {self._rectangle.size_pixels_width=},"
            f" {self.zarr_path=}"
        )

    def open(self) -> None:
        """
        Open NWP data

        We don't want to open_nwp() in __init__.
        If we did that, then we couldn't copy NWPDataSource
        instances into separate processes.  Instead,
        call open() _after_ creating separate processes.
        """
        self._data = self._open_data()
        np.testing.assert_array_equal(self._data["variable"].values, self.channels)
        self._data = self._data.rename({"x": "x_osgb", "y": "y_osgb"})

    def _open_data(self) -> xr.DataArray:
        data = open_nwp(self.zarr_path, consolidated=self.consolidated)
        return data.sel(variable=list(self.channels))

    @staticmethod
    def get_data_model_for_batch():
        """Get the model that is used in the batch"""
        return NWP

    def _get_time_slice(self, t0_datetime_utc: pd.Timestamp) -> xr.DataArray:
        """
        Select the numerical weather predictions for a single time slice.

        Note that this function does *not* resample from hourly to 5 minutely.
        Resampling would be very expensive if done on the whole geographical
        extent of the NWP data!

        Args:
            t0_datetime_utc: the time slice is around t0_dt.

        Returns: Slice of data

        """
        start_dt = self._get_start_dt(t0_datetime_utc)
        end_dt = self._get_end_dt(t0_datetime_utc)

        start_hourly = start_dt.floor("H")
        end_hourly = end_dt.ceil("H")

        # TODO: Issue #398: Use NWP init time closest to t0.
        init_time_i = np.searchsorted(self.data.init_time, start_hourly.to_numpy(), side="right")
        init_time_i -= 1  # Because searchsorted() gives the index to the entry _after_.
        init_time = self.data.init_time.values[init_time_i]

        step_start = start_hourly - init_time
        step_end = end_hourly - init_time

        selected = self.data.sel(init_time=init_time, step=slice(step_start, step_end))
        selected = selected.swap_dims({"step": "target_time"})
        selected["target_time"] = init_time + selected.step
        return selected

    def _post_process_example(self, selected_data: xr.Dataset, t0_dt: pd.Timestamp) -> xr.Dataset:
        """Resamples to 5 minutely."""

        start_dt = self._get_start_dt(t0_dt)
        end_dt = self._get_end_dt(t0_dt)

        # if t0_dt is not on the hour, e.g. 13.05.
        # Then if the history_minutes is 1 hours,
        # so start_dt will be 12.05, but we want to the 12.00 time step too
        start_dt = start_dt.floor("H")

        selected_data = selected_data.sel(target_time=slice(start_dt, end_dt))
        selected_data = selected_data.rename({"target_time": "time", "variable": "channels"})
        selected_data.data = selected_data.data.astype(np.float32)

        return selected_data

    def datetime_index(self) -> pd.DatetimeIndex:
        """Returns a complete list of all available datetimes"""
        if self._data is None:
            nwp = self._open_data()
        else:
            nwp = self._data

        # We need to return the `target_times` (the times the NWPs are _about_).
        # The `target_time` is the `init_time` plus the forecast horizon `step`.
        # `step` is an array of timedeltas, so we can just add `init_time` to `step`.
        target_times = nwp["init_time"] + nwp["step"]
        target_times = target_times.values.flatten()
        target_times = np.unique(target_times)
        target_times = np.sort(target_times)
        target_times = pd.DatetimeIndex(target_times)

        return target_times

    @property
    def sample_period_minutes(self) -> int:
        """Override the default sample minutes"""
        return 60


def open_nwp(zarr_path: str, consolidated: bool) -> xr.DataArray:
    """
    Open The NWP data

    Args:
        zarr_path: zarr_path must start with 'gs://' if it's on GCP.
        consolidated: Is the Zarr metadata consolidated?

    Returns: NWP data.
    """
    _LOG.debug("Opening NWP data: %s", zarr_path)
    utils.set_fsspec_for_multiprocess()

    if zarr_path.split(".")[-1] == "netcdf":
        # loading netcdf file, download bytes and then load as xarray
        with fsspec.open(zarr_path, mode="rb") as file:
            file_bytes = file.read()

        with io.BytesIO(file_bytes) as file:
            nwp = xr.load_dataset(file, engine="h5netcdf")

    else:
        nwp = xr.open_dataset(
            zarr_path,
            engine="zarr",
            consolidated=consolidated,
            mode="r",
            chunks=None,  # Reading Satellite Zarr benefits from setting chunks='auto'
            # (see issue #456) but 'auto' massively slows down reading NWPs.
        )

    # Select the "UKV" DataArray from the "nwp" Dataset.
    # "UKV" is the one and only DataArray in the Zarr Dataset.
    # "UKV" stands for "United Kingdom Variable", and it the UK Met Office's high-res deterministic
    # NWP for the UK.  All the NWP variables are represented in the `variable` dimension within
    # the UKV DataArray.
    ukv = nwp["UKV"]

    # Reverse `y` so it's top-to-bottom (so ZarrDataSource.get_example() works correctly!)
    # if necessary.  Adapted from:
    # https://stackoverflow.com/questions/54677161/xarray-reverse-an-array-along-one-coordinate
    if ukv.y[0] < ukv.y[1]:
        _LOG.warning(
            "NWP y axis runs from bottom-to-top.  Will reverse y axis so it runs top-to-bottom."
        )
        y_reversed = ukv.y[::-1]
        ukv = ukv.reindex(y=y_reversed)

    # Sanity checks.
    ukv = utils.drop_duplicate_times(data_array=ukv, class_name="NWP", time_dim="init_time")

    # If any init_times are not monotonic_increasing then drop the out-of-order init_times:
    ukv = utils.drop_non_monotonic_increasing(
        data_array=ukv, class_name="NWP", time_dim="init_time"
    )

    assert pd.DatetimeIndex(ukv["init_time"]).is_unique
    assert pd.DatetimeIndex(ukv["init_time"]).is_monotonic_increasing

    return ukv



""" Model for output of general/metadata data, useful for a batch """

import logging
from pathlib import Path
from typing import List, Optional, Union

import pandas as pd
from pydantic import BaseModel, Field, validator

from nowcasting_dataset.consts import SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME
from nowcasting_dataset.filesystem.utils import check_path_exists
from nowcasting_dataset.utils import get_start_and_end_example_index

logger = logging.getLogger(__name__)


class SpaceTimeLocation(BaseModel):
    """Location of the example"""

    t0_datetime_utc: pd.Timestamp = Field(
        ...,
        description="The t0 of one example ",
    )

    x_center_osgb: float = Field(
        ...,
        description="The x center of one example in OSGB coordinates",
    )

    y_center_osgb: float = Field(
        ...,
        description="The y center of one example in OSGB coordinates",
    )

    id: Optional[int] = Field(
        None,
        description="The id of the GSP or the PV system. This is optional so can be None",
    )

    id_type: Optional[str] = Field(
        None,
        description="The type of the id. Should be either None, 'gsp' or 'pv_system'",
    )

    @validator("t0_datetime_utc")
    def v_t0_datetime_utc(cls, t0_datetime_utc):
        """Make sure t0_datetime_utc is pandas Timestamp"""
        return pd.Timestamp(t0_datetime_utc)

    @validator("id_type")
    def v_id_type(cls, id_type):
        """Make sure id_type is either None, 'gsp' or 'pv_system'"""

        if id_type == "None":
            id_type = None

        assert id_type in [
            None,
            "gsp",
            "pv_system",
        ], f"{id_type=} should be None, 'gsp' or 'pv_system'"
        return id_type


class Metadata(BaseModel):
    """Class to store metadata data"""

    batch_size: int = Field(
        ...,
        g=0,
        description="The size of this batch. If the batch size is 0, "
        "then this item stores one data item",
    )

    space_time_locations: List[SpaceTimeLocation]

    @property
    def t0_datetimes_utc(self) -> list:
        """Return all the t0"""
        return [location.t0_datetime_utc for location in self.space_time_locations]

    @property
    def x_centers_osgb(self) -> List[float]:
        """List of all the x centers from all the locations"""
        return [location.x_center_osgb for location in self.space_time_locations]

    @property
    def y_centers_osgb(self) -> List[float]:
        """List of all the x centers from all the locations"""
        return [location.y_center_osgb for location in self.space_time_locations]

    @property
    def ids(self) -> List[float]:
        """List of all the ids from all the locations"""
        return [location.id for location in self.space_time_locations]

    def save_to_csv(self, path):
        """
        Save metadata to a csv file

        Args:
            path: the path where the file should be save

        """

        filename = f"{path}/{SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME}"
        metadata_dict = [location.dict() for location in self.space_time_locations]
        # metadata_dict.pop("batch_size")

        # if file exists, add to it
        try:
            check_path_exists(filename)
        except FileNotFoundError:
            metadata_df = pd.DataFrame(metadata_dict)

        else:

            metadata_df = pd.read_csv(filename)

            metadata_df_extra = pd.DataFrame(metadata_dict)
            metadata_df = metadata_df.append(metadata_df_extra)

        metadata_df.to_csv(filename, index=False)


def load_from_csv(
    path: Union[str, Path], batch_size: int, batch_idx: Optional[int] = None
) -> Metadata:
    """
    Load metadata from csv

    Args:
        path: the path which stores the metadata file
        batch_idx: the batch index
        batch_size: how many examples in each batch

    Returns: Metadata class
    """
    filename = f"{path}/{SPATIAL_AND_TEMPORAL_LOCATIONS_OF_EACH_EXAMPLE_FILENAME}"

    # get start and end example index
    if batch_idx is not None:
        start_example_idx, end_example_idx = get_start_and_end_example_index(
            batch_idx=batch_idx, batch_size=batch_size
        )
        skiprows = start_example_idx + 1  # p+1 is to ignore header
        nrows = batch_size
    else:
        skiprows = 1  # ignore header
        nrows = None

    names = list(SpaceTimeLocation.__fields__)

    # read the file
    # kswargs = {}
    # if (start_example_idx is not None) and (end_example_idx is not None):
    #     kswargs['nrows'] = batch_size
    metadata_df = pd.read_csv(
        filename,
        skiprows=skiprows,
        nrows=nrows,
        names=names,
    )
    # If the CSV doesn't contain the optional columns, then drop those missing columns.
    # (Otherwise Pandas creates a column with NaNs, which confuses Pydantic!)
    metadata_df = metadata_df.dropna(axis="columns", how="all")

    assert (
        len(metadata_df) > 0
    ), f"Could not load metadata for {batch_size=} {batch_idx=} {filename=}"

    if "id_type" in metadata_df.columns:
        metadata_df["id_type"].fillna("None", inplace=True)

    # add batch_size
    locations_dict = metadata_df.to_dict("records")
    metadata_dict = {"space_time_locations": locations_dict, "batch_size": batch_size}

    try:
        metadata = Metadata(**metadata_dict)
    except Exception as e:
        logger.error(f"Try to make Metadata object from {metadata_dict}")
        raise e

    return metadata
